# coding: utf-8

"""

    Cancer Imaging Archive Interface
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import io
import os
import pickle
import shutil
import zipfile
import requests
import numpy as np
import pandas as pd
from PIL import Image
from time import sleep
from copy import deepcopy
from warnings import warn
from datetime import datetime
from math import log10, floor

from biovida import __version_numeric__

# General Image Support Tools
from biovida.images._image_tools import NoResultsFound

# Database Management
from biovida.images.image_cache_mgmt import (_records_db_merge,
                                             _record_update_dbs_joiner,
                                             _prune_rows_with_deleted_images)

# General Support Tools
from biovida.support_tools.support_tools import (cln,
                                                 tqdm,
                                                 dicom,
                                                 header,
                                                 items_null,
                                                 only_numeric,
                                                 combine_dicts,
                                                 camel_to_snake_case,
                                                 data_frame_col_drop,
                                                 list_to_bulletpoints,
                                                 IN_NOTEBOOK)

# Import Printing Tools
from biovida.support_tools.printing import pandas_pprint

# Cache Management
from biovida.support_tools._cache_management import package_cache_creator

# Interface Support tools
from biovida.images._interface_support.shared import save_records_db
from biovida.images._interface_support.dicom_data_to_dict import dicom_to_dict
from biovida.images._interface_support.cancer_image.cancer_image_parameters import CancerImageArchiveParams
from biovida.images._interface_support.cancer_image.cancer_image_support_tools import nonessential_cancer_image_columns


# ----------------------------------------------------------------------------------------------------------
# Summarize Studies Provided Through the Cancer Imaging Archive
# ----------------------------------------------------------------------------------------------------------


class _CancerImageArchiveOverview(object):
    """

    Overview of Information Available on the Cancer Imaging Archive.

    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                        one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str`` or ``None``
    :param verbose: if ``True`` print additional information. Defaults to ``False``.
    :type verbose: ``bool``
    :param tcia_homepage: URL to the the Cancer Imaging Archive's homepage.
    :type tcia_homepage: ``str``
    """
    # ToDo: add link to the TCIA page for a given collection/study (use BeautifulSoup).
    def __init__(self,
                 dicom_modality_abbrevs,
                 verbose=False,
                 cache_path=None,
                 tcia_homepage='http://www.cancerimagingarchive.net'):
        self._verbose = verbose
        self._tcia_homepage = tcia_homepage
        _, self._created_image_dirs = package_cache_creator(sub_dir='images', cache_path=cache_path,
                                                            to_create=['tcia'], nest=[('tcia', 'databases')],
                                                            verbose=verbose)

        self.dicom_modality_abbrevs = dicom_modality_abbrevs

    def _all_studies_parser(self):
        """

        Get a record of all studies on the Cancer Imaging Archive.

        :return: the table on the homepage
        :rtype: ``Pandas DataFrame``
        """
        # Extract the main summary table from the home page
        summary_df = pd.read_html(str(requests.get(self._tcia_homepage).text), header=0)[0]

        # Convert column names from camelCase to snake_cake
        summary_df.columns = list(map(camel_to_snake_case, summary_df.columns))

        summary_df = summary_df[summary_df['status'].str.strip().str.lower() != 'coming soon']
        summary_df = summary_df[~summary_df['location'].str.lower().str.contains('phantom')]
        summary_df = summary_df[~summary_df['collection'].str.lower().str.contains('mouse|phantom')]
        summary_df = summary_df[summary_df['access'].str.strip().str.lower() == 'public'].reset_index(drop=True)

        # Add Full Name for Modalities
        summary_df['modalities_full'] = summary_df['modalities'].map(
            lambda x: [self.dicom_modality_abbrevs.get(cln(i), i) for i in cln(x).split(", ")])

        # Parse the Location Column (and account for special case: 'Head-Neck').
        summary_df['location'] = summary_df['location'].map(
            lambda x: cln(x.replace(" and ", ", ").replace("Head-Neck", "Head, Neck")).split(", "))

        summary_df['updated'] = pd.to_datetime(summary_df['updated'], infer_datetime_format=True)
        summary_df.columns = list(map(lambda x: cln(x, extent=2), summary_df.columns))

        return summary_df

    def _obtain_tcia_overview(self, download_override=False):
        """

        Obtain and Manage a copy the table which summarizes the the Cancer Imaging Archive
        on the organization's homepage.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :return: summary table hosted on the home page of the Cancer Imaging Archive.
        :rtype: ``Pandas DataFrame``
        """
        # Define the path to save the data
        save_path = os.path.join(self._created_image_dirs['databases'], 'all_tcia_studies.p')

        if not os.path.isfile(save_path) or download_override:
            if self._verbose:
                header("Downloading Table of Available Studies... ", flank=False)
            summary_df = self._all_studies_parser()
            summary_df.to_pickle(save_path)
        else:
            summary_df = pd.read_pickle(save_path)

        return summary_df

    @staticmethod
    def _studies_filter(summary_df, cancer_type, location, modality):
        """

        Apply Filters passed to ``studies()``.

        :param summary_df: see: ``studies()``.
        :type summary_df: ``Pandas DataFrame``
        :param cancer_type: see: ``studies()``.
        :type cancer_type: ``str``, ``iterable`` or ``None``
        :param location: see: ``studies()``.
        :type location: ``str``, ``iterable`` or ``None``
        :param modality: see: ``studies()``.
        :type modality: ``str``, ``iterable`` or ``None``
        :return: ``summary_df`` with filters applied.
        :type: ``Pandas DataFrame``
        """
        # Filter by `cancer_type`
        if isinstance(cancer_type, (str, list, tuple)):
            if isinstance(cancer_type, (list, tuple)):
                cancer_type = "|".join(map(lambda x: cln(x).lower(), cancer_type))
            else:
                cancer_type = cln(cancer_type).lower()
            summary_df = summary_df[summary_df['cancer_type'].str.lower().str.contains(cancer_type)]

        # Filter by `location`
        if isinstance(location, (str, list, tuple)):
            location = [location] if isinstance(location, str) else location
            summary_df = summary_df[summary_df['location'].map(
                lambda x: any([cln(l).lower() in i.lower() for i in x for l in location]))]

        def modality_filter(x, modality):
            """Apply filter to look for rows which match `modality`."""
            sdf_modalities = cln(x['modalities']).lower()
            sdf_modalities_full = [cln(i).lower() for i in x['modalities_full']]

            if any(m in sdf_modalities for m in modality):
                return True
            if any([m in m_full for m in modality for m_full in sdf_modalities_full]):
                return True
            else:
                return False

        # Filter by `modality`.
        if isinstance(modality, (str, list, tuple)):
            modality = [modality.lower()] if isinstance(modality, str) else list(map(lambda x: x.lower(), modality))
            summary_df = summary_df[summary_df.apply(lambda x: modality_filter(x, modality), axis=1)]

        return summary_df


# ----------------------------------------------------------------------------------------------------------
# Pull Records from the Cancer Imaging Archive
# ----------------------------------------------------------------------------------------------------------


class _CancerImageArchiveRecords(object):
    """

    Class to harvest records for a given collection/study through the Cancer Imaging Archive API.

    :param api_key: an key to the the Cancer Imaging Archive's API.
    :type api_key: ``str``
    :param dicom_modality_abbrevs: an instance of ``CancerImageArchiveParams().dicom_modality_abbreviations('dict')``
    :type dicom_modality_abbrevs: ``dict``
    :param tcia_overview_df: yield of ``_CancerImageArchiveOverview()._obtain_tcia_overview()``
    :type tcia_overview_df: ``Pandas DataFrames``
    :param root_url: the root URL for the Cancer Imaging Archive's API.
    :type root_url: ``str``
    """

    def __init__(self, api_key, dicom_modality_abbrevs, tcia_overview_df, root_url):
        self.ROOT_URL = root_url
        self.records_df = None
        self.dicom_modality_abbrevs = dicom_modality_abbrevs
        self._tcia_overview_df = tcia_overview_df
        self.API_KEY = api_key
        self._url_sep = '+'

    def _study_extract(self, study):
        """

        Download all patients in a given study.

        :param study: a Cancer Imaging Archive collection (study)
        :type study: ``str``
        :return: the yield of passing the ``getPatientStudy`` parameter to the Cancer Imaging Archive API for a given
                 collection (study).
        :rtype: ``Pandas DataFrame``
        """
        url = '{0}/query/getPatientStudy?Collection={1}&format=csv&api_key={2}'.format(
            self.ROOT_URL, cln(study).replace(' ', self._url_sep), self.API_KEY)
        data_frame = pd.DataFrame.from_csv(url).reset_index()

        # Convert column names from camelCase to snake_cake
        data_frame.columns = list(map(camel_to_snake_case, data_frame.columns))

        return data_frame

    def _robust_study_extract(self, study):
        """

        This method uses '+' first as a replacement for spaces when sending requests to the Cancer Imaging Archive.
        If that fails, this method falls back on '-'.

        :param study: a Cancer Imaging Archive collection (study).
        :type study: ``str``
        :return: see: ``_study_extract()``
        :rtype: ``Pandas DataFrame``

        :raises IndexError: if both '+' and '-' fail to yield a dataframe with nonzero length.
        """
        study_df = self._study_extract(study)
        if study_df.shape[0] == 0:
            self._url_sep = '-'
            study_df = self._study_extract(study)
            self._url_sep = '+'  # reset
            if study_df.shape[0] == 0:
                raise IndexError("The '{0}' collection/study data has no length.\n"
                                 "This is likely the result of a problem with the "
                                 "Cancer Imaging Archive REST API's.\n".format(study))
        return study_df

    @staticmethod
    def _date_index_map(list_of_dates):
        """

        Returns a dict of the form: ``{date: index in ``list_of_dates``, ...}``

        :param list_of_dates: a list (or tuple) of datetime objects.
        :type list_of_dates: ``list`` or ``tuple``
        :return: description (above)
        :rtype: ``dict``
        """
        return {k: i for i, k in enumerate(sorted(list_of_dates), start=1)}

    def _summarize_study_by_patient(self, study):
        """

        Summarizes a study by patient.

        :param study: a Cancer Imaging Archive collection (study).
        :type study: ``str``
        :return: nested dictionary of the form:

                ``{patient_id: {study_instance_uid: {'sex': ..., 'age': ..., 'session': ..., 'study_date': ...}}}``

        :rtype: ``dict``
        """
        # Download a summary of all patients in a study
        study_df = self._robust_study_extract(study)

        study_df['study_date'] = pd.to_datetime(study_df['study_date'], infer_datetime_format=True)

        # Divide Study into stages (e.g., Baseline (session 1); Baseline + 1 Month (session 2), etc.
        stages = study_df.groupby('patient_id').apply(
            lambda x: self._date_index_map(x['study_date'].tolist())).to_dict()

        # Apply stages
        study_df['session'] = study_df.apply(lambda x: stages[x['patient_id']][x['study_date']], axis=1)

        # Define columns to extract from `study_df`
        valuable_cols = ('patient_id', 'study_instance_uid', 'session', 'patient_sex', 'patient_age', 'study_date')

        # Convert to a nested dictionary
        patient_dict = dict()
        for pid, si_uid, session, sex, age, date in zip(*[study_df[c] for c in valuable_cols]):
            inner_nest = {'sex': sex, 'age': age, 'session': session, 'study_date': date}
            if pid not in patient_dict:
                patient_dict[pid] = {si_uid: inner_nest}
            else:
                patient_dict[pid] = combine_dicts(patient_dict[pid], {si_uid: inner_nest})

        return patient_dict

    def _patient_image_summary(self, patient, study, patient_dict):
        """

        Harvests the Cancer Image Archive's Text Record of all baseline images for a given patient
        in a given study.

        :param patient: the patient_id (will be used to form the request to the TCIA server).
        :type patient: ``str``
        :param study: a Cancer Imaging Archive collection (study).
        :type study: ``str``
        :param patient_dict: a value in ``study_dict`` (which is a dictionary itself).
        :type patient_dict: ``dict``
        :return: the yield of the TCIA ``getSeries`` param for a given patient in a given collection (study).
                 Their sex, age, the session number (e.g., baseline = 1, baseline + 1 month = 2, etc.) and the
                 'study_date' (i.e., the date the study was conducted).
        :rtype: ``Pandas DataFrame``
        """
        # Select an individual Patient
        url = '{0}/query/getSeries?Collection={1}&PatientID={2}&format=csv&api_key={3}'.format(
            self.ROOT_URL, cln(study).replace(' ', self._url_sep), patient, self.API_KEY)
        patient_df = pd.DataFrame.from_csv(url).reset_index()

        # Convert column names from camelCase to snake_cake
        patient_df.columns = list(map(camel_to_snake_case, patient_df.columns))

        # Add sex, age, session, study_date and patient id
        patient_info = patient_df['study_instance_uid'].map(
            lambda x: {k: patient_dict[x][k] for k in ('sex', 'age', 'session', 'study_date')})
        patient_df = patient_df.join(pd.DataFrame(patient_info.tolist()))
        patient_df['patient_id'] = patient

        return patient_df

    def _clean_patient_study_df(self, patient_study_df):
        """

        Cleans the input in the following ways:

            - convert 'F' --> 'Female' and 'M' --> 'Male'

            - Converts the 'age' column to numeric (years)

            - Remove line breaks in the 'protocol_name' and 'series_description' columns

            - Add Full name for modality (modality_full)

            - Convert the 'series_date' column to datetime

        :param patient_study_df: the ``patient_study_df`` dataframe evolved inside ``_pull_records()``.
        :type patient_study_df: ``Pandas DataFrame``
        :return: a cleaned ``patient_study_df``
        :rtype: ``Pandas DataFrame``
        """
        # convert 'F' --> 'female' and 'M' --> 'male'.
        patient_study_df['sex'] = patient_study_df['sex'].map(
            lambda x: {'F': 'female', 'M': 'male'}.get(cln(str(x)).upper(), x),
            na_action='ignore')

        # Convert entries in the 'age' Column to floats.
        patient_study_df['age'] = patient_study_df['age'].map(
            lambda x: only_numeric(x) / 12.0 if 'M' in str(x).upper() else only_numeric(x),
            na_action='ignore')

        # Remove unneeded line break marker
        for c in ('protocol_name', 'series_description', 'manufacturer_model_name'):
            patient_study_df[c] = patient_study_df[c].map(
                lambda x: cln(x.replace("\/", " ")) if isinstance(x, str) else x, na_action='ignore')

        # Add the full name for modality.
        patient_study_df['modality_full'] = patient_study_df['modality'].map(
            lambda x: self.dicom_modality_abbrevs.get(x, np.NaN), na_action='ignore')

        # Lower and clean 'body_part_examined' column
        patient_study_df['body_part_examined'] = patient_study_df['body_part_examined'].map(
            # ToDo: generalize the special case used here to handle 'headneck'.
            lambda x: cln(x).lower().replace("headneck", "head, neck") if isinstance(x, str) else x,
            na_action='ignore')

        patient_study_df['series_date'] = pd.to_datetime(patient_study_df['series_date'],
                                                         infer_datetime_format=True)

        def series_number_rescale(x):
            if isinstance(x, (int, float)) and not items_null(x) and x > 0:
                return x / 10**floor(log10(x))
            else:
                return x

        patient_study_df['series_number_rescaled'] = patient_study_df['series_number'].map(
            series_number_rescale, na_action='ignore')

        sort_by = ['patient_id', 'session', 'series_number_rescaled']
        return patient_study_df.fillna(np.NaN).sort_values(by=sort_by).reset_index(drop=True)

    def _get_condition_name(self, collection_series):
        """

        This method gets the name of the condition studied for a given collection (study).
        (collection_series gets reduced down to a single unique).

        :param collection_series: a series of the study name, e.g., ('MY-STUDY', 'MY-STUDY', 'MY-STUDY', ...).
        :type collection_series: ``Pandas Series``
        :return: the name of disease studied in a given collection (lower case).
        :rtype: ``str``
        """
        unique_studies = collection_series.unique()
        if len(unique_studies) == 1:
            collection = unique_studies[0]
        else:
            raise AttributeError("`{0}` studies found in `records`. "
                                 "Expected one.".format(str(len(unique_studies))))

        condition_name = self._tcia_overview_df[
            self._tcia_overview_df['collection'] == collection]['cancer_type'].iloc[0]

        return condition_name.lower() if isinstance(condition_name, str) else condition_name

    def records_pull(self, study, search_dict, pull_time, patient_limit, verbose):
        """

        Extract record of all images for all patients in a given study.

        :param study: a Cancer Imaging Archive collection (study).
        :type study: ``str``
        :param search_dict: a dictionary which contains the search information provided by the user
                            (as evolved inside  ``CancerImageInterface()_search_dict_gen()``.
        :type search_dict: ``dict``
        :param pull_time: the time the query was launched.
        :type pull_time: ``datetime``
        :param patient_limit: limit on the number of patients to extract.
                              Patient IDs are sorted prior to this limit being imposed.
                              If ``None``, no `patient_limit` will be imposed.
        :type patient_limit: ``int`` or ``None``
        :param verbose: if ``True`` print additional information. Defaults to ``False``.
        :type verbose: ``bool``
        :return: a dataframe of all baseline images
        :rtype: ``Pandas DataFrame``
        """
        study_dict = self._summarize_study_by_patient(study)

        # Check for invalid `patient_limit` values:
        if not isinstance(patient_limit, int) and patient_limit is not None:
            raise ValueError('`patient_limit` must be an integer or `None`.')
        elif isinstance(patient_limit, int) and patient_limit < 1:
            raise ValueError('If `patient_limit` is an integer it must be greater than or equal to 1.')

        # Define number of patients to extract
        s_patients = sorted(study_dict.keys())
        patients_to_obtain = s_patients[:patient_limit] if isinstance(patient_limit, int) else s_patients

        # Evolve a dataframe ('frame') for the baseline images of all patients
        frames = list()
        for patient in tqdm(patients_to_obtain, desc="'{0}' Records".format(study), disable=not verbose):
            frames.append(self._patient_image_summary(patient, study=study, patient_dict=study_dict[patient]))

        # Concatenate baselines frame for each patient
        patient_study_df = pd.concat(frames, ignore_index=True)

        patient_study_df['article_type'] = ['case_report'] * patient_study_df.shape[0]
        patient_study_df['study_name'] = study
        patient_study_df['cancer_type'] = self._get_condition_name(patient_study_df['collection'])

        # Add the Search query which created the current results and the time the search was launched.
        patient_study_df['query'] = [search_dict] * patient_study_df.shape[0]
        patient_study_df['pull_time'] = [pull_time] * patient_study_df.shape[0]

        self.records_df = self._clean_patient_study_df(patient_study_df)

        return self.records_df


# ----------------------------------------------------------------------------------------------------------
# Pull Images from the Cancer Imaging Archive
# ----------------------------------------------------------------------------------------------------------


class _CancerImageArchiveImages(object):
    """

    Class to harvest images for a given collection/study through the Cancer Imaging Archive API, based on
    records extracted by ``_CancerImageArchiveRecords()``.

    :param api_key: an key to the the Cancer Imaging Archive's API.
    :type api_key: ``str``
    :param dicom_modality_abbrevs: an instance of ``CancerImageArchiveParams().dicom_modality_abbreviations('dict')``
    :type dicom_modality_abbrevs: ``dict``
    :param root_url: the root URL for the the Cancer Imaging Archive's API.
    :type root_url: ``str``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    :type cache_path: ``str`` or ``None``
    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                       one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str`` or ``None``
    """

    def __init__(self, api_key, dicom_modality_abbrevs, root_url, verbose, cache_path=None):
        _, self._created_image_dirs = package_cache_creator(sub_dir='images',
                                                            cache_path=cache_path,
                                                            to_create=['tcia'],
                                                            nest=[('tcia', 'raw'), ('tcia', 'dicoms'),
                                                                  ('tcia', 'databases')],
                                                            verbose=verbose)

        self.verbose = verbose
        self.API_KEY = api_key
        self.dicom_modality_abbrevs = dicom_modality_abbrevs
        self.ROOT_URL = root_url

        # Add Record DataFrame; this database updates in real-time as the images are downloaded.
        self.records_db_images = None
        self.real_time_update_db = None

        # Define the path to the temporary directory
        self.temp_directory_path = os.path.join(self._created_image_dirs['databases'], "__temp__")

    def _instantiate_real_time_update_db(self, db_index):
        """

        Create the ``real_time_update_db`` and define the path to the location where it will be saved.

        :param db_index: the index of the ``real_time_update_db`` dataframe (should be from ``records_db``).
        :type db_index: ``Pandas Series``
        """
        real_time_update_columns = ['cached_dicom_images_path', 'cached_images_path', 'error_free_conversion',
                                    'allowed_modality', 'image_count_converted_cache']

        self.real_time_update_db = pd.DataFrame(
            columns=real_time_update_columns, index=db_index).replace({np.NaN: None})

    def _download_zip(self, series_uid, temporary_folder):
        """

        Downloads the zipped from from the Cancer Imaging Archive for a given 'SeriesInstanceUID' (``series_uid``).

        :param series_uid: the 'series_instance_uid' needed to use TCIA's ``getImage`` parameter
        :type series_uid: ``str``
        :param temporary_folder: path to the temporary folder where the images will be (temporary) cached.
        :type temporary_folder: ``str``
        :return: list of paths to the new files.
        :rtype: ``list``
        """
        # See: http://stackoverflow.com/a/14260592/4898004
        url = '{0}/query/getImage?SeriesInstanceUID={1}&format=csv&api_key={2}'.format(
            self.ROOT_URL, series_uid, self.API_KEY)
        r = requests.get(url)
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(temporary_folder)

        def file_path_full(f):
            """Construct the full path for a given file in ``z``."""
            base_name = cln(os.path.basename(f.filename))
            return os.path.join(temporary_folder, base_name) if len(base_name) else None

        # Generate the list of paths to the dicoms
        return list(filter(None, map(file_path_full, z.filelist)))

    @staticmethod
    def _extract_instance_number(f):
        """
        
        Extract the instance number from a DICOM file.
        
        :param f: a DICOM file
        :param f: ``pydicom object``
        :return: instance number
        :rtype: ``str``
        """
        try:
            instance_number = f.InstanceNumber
        except:
            return 'NA'

        cleaned_instance_number = cln(str(instance_number))
        if len(cleaned_instance_number) and cleaned_instance_number.isdigit():
            return cleaned_instance_number
        else:
            return 'NA'

    def _dicom_to_standard_image(self, f, pull_position, series_uid, conversion, new_file_name):
        """

        This method handles the act of saving dicom images as in a more common file format (e.g., .png).
        An image (``f``) can be either 2 or 3 Dimensional.

        Notes:

        - 3D images will be saved as individual frames

        - if pydicom cannot render the DICOM as a pixel array, this method will its halt image extraction efforts.

        :param f: a (py)dicom image.
        :type f: ``pydicom object``
        :param pull_position: the position of the file in the list of files pulled from the database.
        :type pull_position: ``int``
        :param series_uid: the 'series_instance_uid' needed to use TCIA's ``getImage`` parameter
        :type series_uid: ``str``
        :param conversion: the color scale conversion to use, e.g., 'LA' or 'RGB'.
        :type conversion: ``str``
        :param new_file_name: see ``_convert_and_save_dicom``'s ``save_name`` parameter.
        :type new_file_name: ``str``
        :return: tuple of the form: ``(a list of paths to saved images, boolean denoting success)``
        :rtype: ``tuple``
        """
        # Note: f.PatientsWeight will extract weight from the dicom.

        # Define a list to populate with a record of all images saved
        all_save_paths = list()

        # Extract a pixel array from the dicom file.
        try:
            pixel_arr = f.pixel_array
        except (UnboundLocalError, TypeError):
            return [], False

        save_location = self._created_image_dirs['raw']

        def save_path(instance):
            """Define the path to save the image to."""
            head = "{0}_{1}".format(instance, pull_position)
            file_name = "{0}__{1}__default.{2}".format(
                head, os.path.basename(new_file_name), "png")
            return os.path.join(save_location, file_name)

        if pixel_arr.ndim == 2:
            # Define save name by combining the images instance in the set.
            path = save_path(self._extract_instance_number(f))
            Image.fromarray(pixel_arr).convert(conversion).save(path)
            all_save_paths.append(path)
        # If ``f`` is a 3D image (e.g., segmentation dicom files), save each layer as a separate file/image.
        elif pixel_arr.ndim == 3:
            for instance, layer in enumerate(range(pixel_arr.shape[0]), start=1):
                path = save_path(instance)
                Image.fromarray(pixel_arr[layer:layer + 1][0]).convert(conversion).save(path)
                all_save_paths.append(path)
        else:
            warn("\nProblem converting an image for\nthe following series_instance_uid:\n\n{0}\n\nCannot "
                 "stabilize {1} dimensional arrays.\nImages must be 2D or 3D.".format(series_uid, pixel_arr.ndim))
            return [], False

        return all_save_paths, True

    def _update_and_set_list(self, index, column, new, return_replacement_len=False):
        """

        Set = add to ``self.real_time_update_db``.
        Update = update the list (or tuple) already there.

        Note: the 'list' added to ``self.real_time_update_db`` will actually be a tuple.

        :param column: the name of the column in ``self.real_time_update_db`` to be updated.
        :type column: ``str``
        :param new: the new list to be added.
        :type new: ``list`` or ``tuple``
        :return: the length of the replacement, if ``return_replacement_len`` is ``True``
        :rtype: ``int`` or ``None``
        """
        current = self.real_time_update_db.get_value(index, column)

        def cleaner(to_clean):
            """Tool which will ensure the output is a `list`."""
            return list(to_clean) if isinstance(to_clean, (list, tuple)) else []

        # Clean `current` and `new`, combine and generate a replacement candidate
        replacement_candidate = tuple(cleaner(current) + cleaner(new))

        # Generate the replacement
        replacement = replacement_candidate if len(replacement_candidate) else np.NaN

        self.real_time_update_db.set_value(index, column, replacement)

        return len(replacement) if return_replacement_len and isinstance(replacement, (list, tuple)) else None

    def _convert_and_save_dicom(self,
                                path_to_dicom_file,
                                index,
                                pull_position,
                                series_uid,
                                save_name,
                                color=False):
        """

        Save a dicom image as a more common file format.

        :param path_to_dicom_file: path to a dicom image
        :type path_to_dicom_file: ``str``
        :param pull_position: the position of the image in the raw zip data provided by the Cancer Imaging Archive API.
        :type pull_position: ``int``
        :param series_uid: the 'series_instance_uid' needed to use TCIA's ``getImage`` parameter
        :type series_uid: ``str``
        :param index: the row index currently being processed inside of the main loop in ``_pull_images_engine()``.
        :type index: ``int``
        :param save_name: name of the new file (do *NOT* include a file extension).
                          If ``None``, name from ``path_to_dicom_file`` will be conserved.
        :type save_name: ``str``
        :param color: If ``True``, convert the image to RGB before saving. If ``False``, save as a grayscale image.
                      Defaults to ``False``
        :type color: ``bool``
        """
        # Load the DICOM file into RAM
        f = dicom.read_file(path_to_dicom_file)

        # Conversion (needed so the resultant image is not pure black)
        conversion = 'RGB' if color else 'LA'  # note: 'LA' = grayscale.

        if isinstance(save_name, str):
            new_file_name = save_name
        else:
            # Remove the file extension and then extract the base name from the path.
            new_file_name = os.path.basename(os.path.splitext(path_to_dicom_file)[0])

        # Convert the image into a PIL object and save to disk.
        all_save_paths, success = self._dicom_to_standard_image(f=f, pull_position=pull_position,
                                                                series_uid=series_uid, conversion=conversion,
                                                                new_file_name=new_file_name)

        # Update Record
        cfp_len = self._update_and_set_list(index, 'cached_images_path', all_save_paths, return_replacement_len=True)
        self.real_time_update_db.set_value(index, 'image_count_converted_cache', cfp_len)

        # Add record of whether or not the dicom file could be converted to a standard image type
        if self.real_time_update_db.get_value(index, 'error_free_conversion') != False:  # Block False being replaced.
            self.real_time_update_db.set_value(index, 'error_free_conversion', success)

    def _move_dicoms(self, dicom_files, series_abbrev, index):
        """

        Move the dicom source files to ``self._created_image_dirs['dicoms']``.
        Employ to prevent the raw dicom files from being destroyed.

        :param dicom_files: the yield of ``_download_zip()``
        :type dicom_files: ``list``
        :param series_abbrev: as evolved inside ``_pull_images_engine()``.
        :type series_abbrev: ``str``
        :param index: the row index currently being processed inside of the main loop in ``_pull_images_engine()``.
        :type index: ``int``
        """
        new_dicom_paths = list()
        for file in dicom_files:
            # Define a name for the new file by extracting the dicom file name and combining with `series_abbrev`.
            instance_number = self._extract_instance_number(f=dicom.read_file(file))
            file_parsed = list(os.path.splitext(os.path.basename(file)))
            new_dicom_file_name = "{0}_{1}__{2}{3}".format(
                # ToDo: this is redundant. The file is read twice: once to move and once if converting to PNG.
                instance_number, file_parsed[0], series_abbrev, file_parsed[1])

            new_location = os.path.join(self._created_image_dirs['dicoms'], new_dicom_file_name)
            new_dicom_paths.append(new_location)

            if os.path.isfile(new_location):
                os.remove(new_location)
            shutil.move(file, new_location)

        def sort_func(i):
            """Sort `new_dicom_paths` by instance_number."""
            try:
                instance_number = os.path.basename(i).split("_")[0]
                return float(instance_number)
            except:
                return len(new_dicom_paths)  # `len` is O(1)

        sorted_paths = tuple(sorted(new_dicom_paths, key=sort_func))
        self.real_time_update_db.set_value(index, 'cached_dicom_images_path', sorted_paths)

    def _cache_check(self, series_abbrev, n_images_min, save_png, save_dicom):
        """

        Check that caches likely contain that data which would be obtained by downloading it from the database.

        :param series_abbrev: as evolved inside ``_pull_images_engine()``
        :type series_abbrev: ``str``
        :param n_images_min: `image_count` as passed in ``_pull_images_engine()``. Denotes the min. number of images
                              for a given series for the cache to be considered complete (less than this number
                              will trigger an effort to download the corresponding images (or, more specifically,
                              SeriesInstanceUID). Note: This number gives the number of dicoms -- the number of
                              converted images could be larger (i.e., 3D images which can unpack to be many more),
                              but it is not possible to known this without actually downloading and unpacking these
                              images.
        :param save_png: ``True`` signals the user's intention to convert DICOMs to PNG.
        :type save_png: ``bool``
        :param save_dicom: see: ``pull_images()``.
        :type save_dicom: ``bool``
        :type n_images_min: ``int``
        :return: tuple of the form:

        ``(cache likely complete,
           series_abbrev matches in self._created_image_dirs['raw'],
           series_abbrev matches in self._created_image_dirs['dicoms'])``

        :rtype: ``tuple``
        """
        # Check that `self._created_image_dirs['raw']` has files which contain the string `series_abbrev`.
        converted_loc_summary = tuple(sorted([os.path.join(self._created_image_dirs['raw'], f)
                                              for f in os.listdir(self._created_image_dirs['raw'])
                                              if series_abbrev in f]))

        # Check that `self._created_image_dirs['dicoms'])` has files which contain the string `series_abbrev`.
        dicoms_sl_summary = tuple(sorted([os.path.join(self._created_image_dirs['dicoms'], f)
                                          for f in os.listdir(self._created_image_dirs['dicoms'])
                                          if series_abbrev in f]))

        if save_png:
            converted_loc_summary_complete = len(converted_loc_summary) >= n_images_min
        else:
            converted_loc_summary_complete = True  # i.e., the actual answer is irrelevant

        if save_dicom:
            dicoms_sl_summary_complete = len(dicoms_sl_summary) >= n_images_min
        else:
            dicoms_sl_summary_complete = True  # i.e., the actual answer is irrelevant

        cache_complete = converted_loc_summary_complete and dicoms_sl_summary_complete

        return (cache_complete,
                converted_loc_summary if len(converted_loc_summary) else None,
                dicoms_sl_summary if len(dicoms_sl_summary) else None)

    def _create_temp_dicom_dir(self):
        """

        Create a temporary directory for DICOM files immediately following download from
        The Cancer Imaging Archive.

        :return: the full path to the newly created temporary directory.
        :rtype: ``str``
        """
        temp_folder = os.path.join(self._created_image_dirs['dicoms'], '__temp_dicom__')

        if os.path.isdir(temp_folder):
            # To prevent duplicate images being created,
            # destroy this directory if it already exists.
            shutil.rmtree(temp_folder, ignore_errors=True)
        os.makedirs(temp_folder)
        return temp_folder

    @staticmethod
    def _valid_modality(allowed_modalities, modality, modality_full):
        """

        Check if `modality` or `modality_full` contains the modality the user is looking for.

        :param allowed_modalities: see: ``pull_images()``
        :type allowed_modalities: ``list``, ``tuple`` or ``None``.
        :param modality: a single element from the ``modality`` column in ``self.real_time_update_db``.
        :type modality: ``str``
        :param modality_full: a single element from the ``modality_full`` column in ``self.real_time_update_db``.
        :type modality_full: ``str``
        :return: whether or not the image satisfies the modality the user is looking for.
        :rtype: ``bool``
        """
        # Assume True if `allowed_modalities` is left to its default (`None`).
        if allowed_modalities is None:
            return True

        if not isinstance(allowed_modalities, (list, tuple, str)):
            raise ValueError("`allowed_modalities` must be one of `list`, `tuple`, `str`.")

        # Convert `allowed_modalities` to an iterable
        if isinstance(allowed_modalities, str):
            allowed_modalities = [allowed_modalities]

        # Check if any item in `allowed_modalities` is a sublist in `modality` or `modality_full`.
        if any([cln(l).lower() in cln(i).lower() for l in allowed_modalities for i in (modality, modality_full)]):
            return True
        else:
            return False

    def _pull_images_engine(self, save_dicom, allowed_modalities, save_png):
        """

        Tool to coordinate the above machinery for pulling and downloading images (or locating them in the cache).

        :param save_dicom: see: ``pull_images()``.
        :type save_dicom: ``bool``
        :param save_png: see: ``pull_images()``
        :param save_png: ``bool``
        :param allowed_modalities: see: ``pull_images()``
        :type allowed_modalities: ``list``, ``tuple`` or ``None``
        """
        for index, row in tqdm(self.records_db_images.iterrows(), total=len(self.records_db_images),
                               desc='Obtaining Images', disable=not self.verbose):
            # Check if the image should be harvested (or loaded from the cache).
            valid_image_modality = self._valid_modality(allowed_modalities, row['modality'], row['modality_full'])

            # Add whether or not the image was of the modality (or modalities) requested by the user.
            self.real_time_update_db.set_value(index, 'allowed_modality', valid_image_modality)

            # Compose central part of the file name from 'patient_id' and the last ten digits of 'series_instance_uid'
            series_abbrev = "{0}_{1}".format(row['patient_id'], str(row['series_instance_uid'])[-10:])

            # Analyze the cache to determine whether or not downloading the images is needed
            cache_complete, sl_summary, dsl_summary = self._cache_check(series_abbrev=series_abbrev,
                                                                        n_images_min=row['image_count'],
                                                                        save_png=save_png,
                                                                        save_dicom=save_dicom)

            if valid_image_modality and not cache_complete:
                temporary_folder = self._create_temp_dicom_dir()

                # Download the images into a temporary folder.
                dicom_files = self._download_zip(row['series_instance_uid'], temporary_folder=temporary_folder)

                if save_png:
                    for e, f in enumerate(dicom_files, start=1):
                        self._convert_and_save_dicom(path_to_dicom_file=f, index=index, pull_position=e,
                                                     series_uid=row['series_instance_uid'],
                                                     save_name=series_abbrev)

                if save_dicom:
                    self._move_dicoms(dicom_files=dicom_files, series_abbrev=series_abbrev, index=index)

                shutil.rmtree(temporary_folder, ignore_errors=True)
            else:
                self._update_and_set_list(index, 'cached_dicom_images_path', dsl_summary)
                self._update_and_set_list(index, 'cached_images_path', sl_summary)
                if save_png:  # otherwise, leave as None/NaN
                    self.real_time_update_db.set_value(index, 'error_free_conversion', cache_complete)
                converted_image_count = len(sl_summary) if isinstance(sl_summary, (list, tuple)) else None
                self.real_time_update_db.set_value(index, 'image_count_converted_cache', converted_image_count)

    def pull_images(self, records_db, session_limit, save_png, save_dicom, allowed_modalities):
        """

        Pull Images from the Cancer Imaging Archive.

        :param records_db: the yield from ``_CancerImageArchiveRecords().records_pull()``.
        :type records_db: ``Pandas DataFrame``
        :param session_limit: restrict image harvesting to the first ``n`` sessions, where ``n`` is the value passed
                              to this parameter. If ``None``, no limit will be imposed.
        :type session_limit: ``int``
        :param save_png: see ``CancerImageArchive().pull()``.
        :type save_png: ``bool``
        :param save_dicom: if ``True``, save the raw dicom files.
        :type save_dicom: ``bool``
        :param allowed_modalities: limit images downloaded to certain modalities.
                                   See: CancerImageInterface().dicom_modality_abbrevs (use the keys).
                                   Note: 'MRI', 'PET', 'CT' and 'X-Ray' can also be used.
                                   This parameter is not case sensitive.
        :type allowed_modalities: ``list``, ``tuple`` or ``None``
        :return: a dataframe with information about the images cached by this method.
        :rtype: ``Pandas DataFrame``
        """
        # Notes on 'image_count_converted_cache':
        # 1. a column which provides the number of images each SeriesInstanceUID yielded
        # 2. values may be discrepant with the 'image_count' column because 3D images are expanded
        #    into their individual frames when saved to the converted images cache.

        # Create the __temp__ folder for `image_pull_settings.p` if it does not already exist.
        if not os.path.isdir(self.temp_directory_path):
            os.makedirs(self.temp_directory_path)

        settings_path = os.path.join(self.temp_directory_path, "image_pull_settings.p")
        pickle.dump({k: v for k, v in locals().items() if k not in ('self', 'settings_path')},
                    open(settings_path, "wb"))

        if isinstance(session_limit, int):
            if session_limit < 1:
                raise ValueError("`session_limit` must be greater than or equal to 1.")
            self.records_db_images = records_db[records_db['session'].map(
                lambda x: float(x) <= session_limit if pd.notnull(x) else False)].reset_index(drop=True).copy(deep=True)
        else:
            self.records_db_images = records_db.reset_index(drop=True).copy(deep=True)

        # Instantiate `self.real_time_update_db`
        self._instantiate_real_time_update_db(db_index=self.records_db_images.index)

        self._pull_images_engine(save_dicom=save_dicom, allowed_modalities=allowed_modalities,
                                 save_png=save_png)
        self.real_time_update_db = self.real_time_update_db.replace({None: np.NaN})

        return _record_update_dbs_joiner(records_db=self.records_db_images, update_db=self.real_time_update_db)


# ----------------------------------------------------------------------------------------------------------
# Construct Database
# ----------------------------------------------------------------------------------------------------------


class CancerImageInterface(object):
    """

    Python Interface for the `Cancer Imaging Archive <http://www.cancerimagingarchive.net/>`_'s API.

    :param api_key: a key to the the Cancer Imaging Archive's API.

        .. note::

            An API key can be obtained by following the instructions provided `here <https://wiki.cancerimagingarchive.net/display/Public/TCIA+Programmatic+Interface+%28REST+API%29+Usage+Guide/>`_.

    :type api_key: ``str``
    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                       one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str`` or ``None``
    :param verbose: if ``True`` print additional details
    :type verbose: ``bool``
    
    .. warning::
    
        If you wish to use any data obtained from this resource for any form of publication,
        you must follow the citation guidelines provided by the study's authors on the project's
        Cancer Imaging Archive repository page.
    
    .. warning::
    
        Several studies on the Cancer Imaging Archive are subject to publication blockades.
        Therefore, if you intend to publish any findings which use data from this resource, you
        must first check that the studies you have selected are not subject to such restrictions.
    
    """

    def _save_cache_records_db(self):
        """

        Save ``cache_records_db`` to 'disk'.

        """
        self.cache_records_db.to_pickle(self._cache_records_db_save_path)

    def _load_prune_cache_records_db(self, load):
        """

        Load and Prune the ``cache_records_db``.

        :param load: if ``True`` load the ``cache_records_db`` dataframe in from disk.
        :type load: ``bool``
        """
        cache_records_db = pd.read_pickle(self._cache_records_db_save_path) if load else self.cache_records_db
        self.cache_records_db = _prune_rows_with_deleted_images(cache_records_db=cache_records_db,
                                                                columns=['cached_images_path',
                                                                         'cached_dicom_images_path'],
                                                                save_path=self._cache_records_db_save_path)

        # Recompute the image_count_converted_cache column following the pruning procedure.
        self.cache_records_db['image_count_converted_cache'] = self.cache_records_db['cached_images_path'].map(
            lambda x: len(x) if isinstance(x, (list, tuple)) else np.NaN)

    def _latent_temp_dir(self):
        """

        Merge any latent elements in the '__temp__' folder, if such a folder exists (and if it is populated).

        """
        if os.path.isdir(self._Images.temp_directory_path):
            temp_dir = self._Images.temp_directory_path
            latent_pickles = [os.path.join(temp_dir, i) for i in os.listdir(temp_dir) if i.endswith(".p")]
            if len(latent_pickles):
                with open(latent_pickles[0], "rb") as f:
                    settings_dict = pickle.load(f)
                settings_dict_for_pull = {k: v for k, v in settings_dict.items() if k not in ['records_db']}
                settings_dict_for_pull['new_records_pull'] = False  # adding this separately in `pull` breaks in python2

                print("\nResuming Download...")
                self.load_records_db(records_db=settings_dict['records_db'])
                self.pull(**settings_dict_for_pull)

    def _tcia_cache_records_db_handler(self):
        """

        Generate the `cache_records_db` database.
        If it does not exist, use `records_db`.
        If if already exists, merge it with `records_db_update`.

        """
        columns_to_drop = ['allowed_modality']

        def rows_to_conserve_func(x):
            """Mark to conserve the row in the cache if the conversion was successful or DICOMs were saved."""
            iccc, cdip = x['image_count_converted_cache'], x['cached_dicom_images_path']
            any_conversion_success = isinstance(iccc, (int, float)) and iccc > 0
            raw_dicoms = isinstance(cdip, (list, tuple)) and len(cdip) > 0  # > 0 needed b/c `True and 37 == 37`.
            return any_conversion_success or raw_dicoms

        def tcia_duplicates_handler(data_frame):
            """ Drop duplicate rows s.t. rows which reference 
            the most number of images are prefered."""
            # Cache_score is more important than keeping the most recent pull.
            sortby = ['biovida_version', '__image_cache_score__', 'pull_time']
            duplicates = ['series_instance_uid', 'study_instance_uid', 'query']

            def image_cache_score(row):
                def images_len(i):
                    return len(i) if isinstance(i, (list, tuple)) else 0
                return images_len(row['cached_images_path']) + images_len(row['cached_dicom_images_path'])

            data_frame['__image_cache_score__'] = data_frame.apply(image_cache_score, axis=1)
            data_frame = data_frame.sort_values(by=sortby)
            data_frame = data_frame.drop_duplicates(subset=duplicates, keep='last')
            return data_frame.drop('__image_cache_score__', axis=1)

        # Compose or update the master 'cache_records_db' dataframe
        if self.cache_records_db is None:
            cache_records_db = self.records_db.copy(deep=True)
            self.cache_records_db = cache_records_db[
                cache_records_db.apply(rows_to_conserve_func, axis=1)
            ].drop(columns_to_drop, axis=1).reset_index(drop=True)
            self._save_cache_records_db()
        else:
            columns_with_iterables_to_sort = ('cached_images_path', 'cached_dicom_images_path')
            self.cache_records_db = _records_db_merge(interface_name='CancerImageInterface',
                                                      current_records_db=self.cache_records_db,
                                                      records_db_update=self.records_db.drop(columns_to_drop, axis=1),
                                                      columns_with_dicts=('query',),
                                                      duplicates=tcia_duplicates_handler,
                                                      rows_to_conserve_func=rows_to_conserve_func,
                                                      columns_with_iterables_to_sort=columns_with_iterables_to_sort)

            # Save to disk
            self._save_cache_records_db()

        # Delete the '__temp__' folder
        shutil.rmtree(self._Images.temp_directory_path, ignore_errors=True)

    def __init__(self, api_key, cache_path=None, verbose=True):
        if os.path.isdir(api_key):
            raise ValueError("`api_key` must be not be a system path.")
        self._API_KEY = api_key
        self._cache_path = cache_path
        self._verbose = verbose
        self.dicom_modality_abbrevs = CancerImageArchiveParams(cache_path, verbose).dicom_modality_abbreviations('dict')

        # Root URL to for the Cancer Imaging Archive's REST API
        root_url = 'https://services.cancerimagingarchive.net/services/v3/TCIA'

        # Instantiate Classes
        self._Overview = _CancerImageArchiveOverview(dicom_modality_abbrevs=self.dicom_modality_abbrevs,
                                                     verbose=verbose,
                                                     cache_path=cache_path)

        self._tcia_overview_df = self._Overview._obtain_tcia_overview()

        self._Records = _CancerImageArchiveRecords(api_key=api_key,
                                                   dicom_modality_abbrevs=self.dicom_modality_abbrevs,
                                                   tcia_overview_df=deepcopy(self._tcia_overview_df),
                                                   root_url=root_url)

        self._Images = _CancerImageArchiveImages(api_key=api_key,
                                                 dicom_modality_abbrevs=self.dicom_modality_abbrevs,
                                                 root_url=root_url,
                                                 cache_path=cache_path,
                                                 verbose=verbose)

        self._ROOT_PATH = self._Overview._created_image_dirs['ROOT_PATH']

        # Search attributes
        self._pull_time = None
        self.search_dict = None
        self.current_query = None

        # Databases
        self.records_db = None
        self.pull_success = None

        # Path to the `cache_records_db`
        self._cache_records_db_save_path = os.path.join(self._Images._created_image_dirs['databases'],
                                                        'tcia_cache_records_db.p')

        # Load `cache_records_db` if it exists already, else set to None.
        if os.path.isfile(self._cache_records_db_save_path):
            self._load_prune_cache_records_db(load=True)
        else:
            self.cache_records_db = None

        # Load in databases in 'databases/__temp__', if they exist
        self._latent_temp_dir()

    def save_records_db(self, path):
        """

        Save the current ``records_db``.

        :param path: a system path ending with the '.p' file extension.
        :type path: ``str``
        """
        save_records_db(data_frame=self.records_db, path=path)

    def load_records_db(self, records_db):
        """

        Load a ``records_db``

        :param records_db: a system path or ``records_db`` itself.
        :type records_db: ``str`` or ``Pandas DataFrame``
        """
        if isinstance(records_db, pd.DataFrame):
            self.records_db = records_db
        else:
            self.records_db = pd.read_pickle(records_db)
        self._pull_time = self.records_db['pull_time'].iloc[0]
        last_query = self.records_db['query'].iloc[0]
        last_query['download_override'] = False
        last_query['pretty_print'] = False
        self.search(**last_query)

    @property
    def records_db_short(self):
        """Return `records_db` with nonessential columns removed."""
        return data_frame_col_drop(self.records_db, nonessential_cancer_image_columns, 'records_db')

    @property
    def cache_records_db_short(self):
        """Return `cache_records_db` with nonessential columns removed."""
        return data_frame_col_drop(self.cache_records_db, nonessential_cancer_image_columns, 'cache_records_db')
    
    def update_collections(self):
        """Refresh the list of collections provided by the Cancer Imaging Archive."""
        self._tcia_overview_df = self._Overview._obtain_tcia_overview(download_override=True)
        self._Records._tcia_overview_df = deepcopy(self._tcia_overview_df)

    @staticmethod
    def _collection_filter(summary_df, collection, cancer_type, location):
        """

        Limits `summary_df` to individual collections.

        :param summary_df: the yield of ``_CancerImageArchiveOverview()._obtain_tcia_overview()``
        :type summary_df: ``Pandas DataFrame``
        :param collection: a collection (study), or iterable (e.g., list) of collections,
                           hosted by the Cancer Imaging Archive. Defaults to ``None``.
        :type collection: ``list``, ``tuple``, ``str`` or ``None``
        :param cancer_type: a string or list/tuple of specifying cancer types. Defaults to ``None``.
        :type cancer_type: ``str``, ``iterable`` or ``None``
        :param location: a string or list/tuple of specifying body locations. Defaults to ``None``.
        :type location: ``str``, ``iterable`` or ``None``
        :return: ``summary_df`` limited to individual collections.
        :rtype: ``None`` or ``Pandas DataFrame``
        """
        # Filter by `collection`
        if isinstance(collection, (str, list, tuple)) and any(i is not None for i in (cancer_type, location)):
            raise ValueError("Both `cancer_types` and `location` must be ``None`` if a `collection` name is passed.")
        elif isinstance(collection, (str, list, tuple)):
            coll = [collection] if isinstance(collection, str) else collection
            summary_df = summary_df[summary_df['collection'].str.lower().isin(map(lambda x: cln(x).lower(), coll))]
            if summary_df.shape[0] == 0:
                raise AttributeError("No collection with the name '{0}' could be found.".format(collection))
            else:
                return summary_df.reset_index(drop=True)
        elif collection is not None:
            raise TypeError("'{0}' is an invalid type for `collection`.".format(str(type(collection).__name__)))

    def _search_dict_gen(self, collection, cancer_type, location, modality):
        """

        Generate a dictionary which contains the search information provided by the user.

        :param collection: See: ``search()``
        :type collection: ``str``, ``iterable`` or ``None``
        :param cancer_type: See: ``search()``
        :type cancer_type: ``str``, ``iterable`` or ``None``
        :param location: See: ``search()``
        :type location: ``str``, ``iterable`` or ``None``
        :param modality: See: ``search()``
        :type modality: ``str``, ``iterable`` or ``None``
        """
        # Note: lowered here because this is the behavior upstream.
        sdict = {'collection': collection, 'cancer_type': cancer_type, 'location': location, 'modality': modality}

        def lower_sdict(v):
            if isinstance(v, str):
                return v.lower()
            elif isinstance(v, (list, tuple)):
                return tuple(map(lambda x: x.lower(), v))
            else:
                return v

        self.search_dict = {k: lower_sdict(v) for k, v in sdict.items()}

    def search(self,
               collection=None,
               cancer_type=None,
               location=None,
               modality=None,
               download_override=False,
               pretty_print=True):
        """

        Method to Search for studies on the Cancer Imaging Archive.

        :param collection: a collection (study), or iterable (e.g., list) of collections,
                           hosted by the Cancer Imaging Archive. Defaults to ``None``.
        :type collection: ``list``, ``tuple``, ``str`` or ``None``
        :param cancer_type: a string or list/tuple of specifying cancer types. Defaults to ``None``.
        :type cancer_type: ``str``, ``iterable`` or ``None``
        :param location: a string or list/tuple of specifying body locations. Defaults to ``None``.
        :type location: ``str``, ``iterable`` or ``None``
        :param modality: the type of imaging technology. See: ``CancerImageInterface().dicom_modality_abbrevs`` for
                         valid values. Defaults to ``None``.
        :type modality: ``str``, ``iterable`` or ``None``
        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :param pretty_print: if ``True``, pretty print the search results. Defaults to ``True``.
        :type pretty_print: ``bool``
        :return: a dataframe containing the search results.
        :rtype: ``Pandas DataFrame``

        :Example:

        >>> CancerImageInterface(YOUR_API_KEY_HERE).search(cancer_type='carcinoma', location='head')
        ...
           collection                   cancer_type                          modalities         subjects    location
        0  TCGA-HNSC            Head and Neck Squamous Cell Carcinoma  CT, MR, PT                 164     [Head, Neck]
        1  QIN-HeadNeck         Head and Neck Carcinomas               PT, CT, SR, SEG, RWV       156     [Head, Neck]
              ...                          ...                                  ...               ...         ...

        """
        self._search_dict_gen(collection, cancer_type, location, modality)

        if download_override:
            self.update_collections()
        summary_df = deepcopy(self._tcia_overview_df)

        if collection is not None:
            summary_df = self._collection_filter(summary_df,
                                                 collection=collection,
                                                 cancer_type=cancer_type,
                                                 location=location)
        else:
            summary_df = self._Overview._studies_filter(summary_df,
                                                        cancer_type=cancer_type,
                                                        location=location,
                                                        modality=modality)
            if summary_df.shape[0] == 0:
                raise NoResultsFound("Try Broadening the Search Criteria.")

        self.current_query = summary_df.reset_index(drop=True)

        if pretty_print and not IN_NOTEBOOK:
            pandas_pprint(data=self.current_query, full_cols=True,
                          col_align='left', column_width_limit=750)

        # Warn the user if search criteria have not been applied.
        if all([collection is None, cancer_type is None, location is None, modality is None]):
            sleep(0.25)
            warn("\nSpecific search criteria have not been applied.\n"
                 "If `pull()` is called, *all* collections will be downloaded.\n"
                 "Such a request could yield several terabytes of data.\n"
                 "If you still wish to proceed, consider adjusting `pull()`'s\n"
                 "`patient_limit` and `session_limit` parameters.")

        if not pretty_print or IN_NOTEBOOK:
            return self.current_query

    def _pull_records(self, patient_limit, collections_limit):
        """

        Pull Records from the TCIA API.

        :param patient_limit: limit on the number of patients to extract.
                             Patient IDs are sorted prior to this limit being imposed.
                             If ``None``, no patient_limit will be imposed. Defaults to `3`.
        :type patient_limit: ``int`` or ``None``
        :param collections_limit: limit the number of collections to download. If ``None``, no limit will be applied.
        :type collections_limit: ``int`` or ``None``
        :return: a list of dataframes
        :rtype: ``list``
        """
        pull_success = list()
        if isinstance(collections_limit, int):
            all_collections = self.current_query['collection'][:collections_limit]
        else:
            all_collections = self.current_query['collection']

        # Loop through and download all of the studies
        record_frames = list()
        for collection in all_collections:
            try:
                record_frames.append(self._Records.records_pull(study=collection,
                                                                search_dict=self.search_dict,
                                                                pull_time=self._pull_time,
                                                                patient_limit=patient_limit,
                                                                verbose=self._verbose))
                pull_success.append((True, collection))
            except IndexError as e:
                warn("\nIndexError Encountered: {0}".format(e))
                pull_success.append((False, collection))

        return record_frames, pull_success

    def _records_db_gen(self, patient_limit, collections_limit):
        """
        
        Generate ``records_db`` by refining the output of ``_pull_records()``.
        
        :param patient_limit: see ``pull()``.
        :type patient_limit: ``int`` or ``None``
        :param collections_limit: see ``pull()``.
        :type collections_limit: ``int`` or ``None``
        :return: 
        """
        record_frames, self.pull_success = self._pull_records(patient_limit=patient_limit,
                                                              collections_limit=collections_limit)

        # Check for failures
        download_failures = [collection for (success, collection) in self.pull_success if success is False]

        if len(download_failures) == len(self.pull_success):
            raise IndexError("Data could not be harvested for any of the requested collections.")
        elif len(download_failures):
            warn("\nThe following collections failed to download:\n{0}".format(
                list_to_bulletpoints(download_failures)))

        # Combine all record frames
        records_db = pd.concat(record_frames, ignore_index=True)
        records_db['biovida_version'] = [__version_numeric__] * records_db.shape[0]

        return records_db

    def extract_dicom_data(self, database='records_db', make_hashable=False):
        """

        Extract data from all dicom files referenced in ``records_db`` or ``cache_records_db``.
        Note: this requires that ``save_dicom`` is ``True`` when ``pull()`` is called.

        :param database: the name of the database to use. Must be one of: 'records_db', 'cache_records_db'.
                         Defaults to 'records_db'.
        :type database: ``str``
        :param make_hashable: If ``True`` convert the data extracted to nested tuples.
                              If ``False`` generate nested dictionaries. Defaults to ``False``
        :type make_hashable: ``bool``
        :return: a series of the dicom data with dictionaries of the form ``{path: {DICOM Description: value, ...}, ...}``.
                 If ``make_hashable`` is ``True``, all dictionaries will be converted to ``tuples``.
        :rtype: ``Pandas Series``
        """
        # ToDo: remove need for ``save_dicom=True`` by calling upstream before the temporary dicom files are destroyed.
        if database == 'records_db':
            database_to_use = self.records_db
        elif database == 'cache_records_db':
            database_to_use = self.cache_records_db
        else:
            raise ValueError("`database` must be one of 'records_db', 'cache_records_db'.")

        if not isinstance(database_to_use, pd.DataFrame):
            raise TypeError('`{0}` is not a DataFrame.'.format(database))
        else:
            db = database_to_use.copy(deep=True)

        def dicom_apply(paths):
            """Extract dicom data."""
            if not isinstance(paths, (list, tuple)):
                return paths
            elif not len(paths):
                return paths
            else:
                if make_hashable:
                    return tuple({p: tuple(dicom_to_dict(dicom_file=p).items()) for p in paths}.items())
                else:
                    return {p: dicom_to_dict(dicom_file=p) for p in paths}

        return pd.Series([dicom_apply(p) for p in db['cached_dicom_images_path']], index=db.index)

    def pull(self,
             patient_limit=3,
             session_limit=1,
             collections_limit=None,
             allowed_modalities=None,
             save_dicom=True,
             save_png=False,
             new_records_pull=True):
        """

        Pull (i.e., download) the current search.

        Notes:

        - When ``save_png`` is ``True``, 3D DICOM images are saved as individual frames.

        - PNG file names in the cache adhere to the following format:

            ``[instance, pull_position]__[patient_id_[Last 10 Digits of SeriesInstanceUID]]__[Image Scale ('default')].png``

        - DICOM file names in the cache adhere to the following format:
        
            ``[instance, original_name_in_source_file]__[patient_id_[Last 10 Digits of SeriesInstanceUID]].dcm``
        
        where:

        .. hlist::
            :columns: 1

            * 'instance' denotes the image's position in the 3D image (if applicable and available)
            * 'pull_position' denotes the position of the image in the set returned for the given 'SeriesInstanceUID' by the Cancer Imaging Archive.

        :param patient_limit: limit on the number of patients to extract.
                             Patient IDs are sorted prior to this limit being imposed.
                             If ``None``, no patient_limit will be imposed. Defaults to `3`.
        :type patient_limit: ``int`` or ``None``
        :param session_limit: restrict image harvesting to the first ``n`` imaging sessions (days) for a given patient,
                              where ``n`` is the value passed to this parameter. If ``None``, no limit will be imposed.
                              Defaults to `1`.

                .. warning::

                        Several studies (collections) in the Cancer Imaging Archive database have multiple imaging sessions.
                        Latter sessions may be of patients following interventions, such as surgery, intended to
                        *eliminate* cancerous tissue. For this reason it cannot be assumed that images obtained from
                        non-baseline sessions (i.e., session number > 1) contain signs of disease.

        :type session_limit: ``int``
        :param collections_limit: limit the number of collections to download. If ``None``, no limit will be applied.
                                  Defaults to ``None``.
        :type collections_limit: ``int`` or ``None``
        :param allowed_modalities: limit images downloaded to certain modalities.
                                   See: ``CancerImageInterface(YOUR_API_KEY_HERE).dicom_modality_abbrevs`` (use the keys).
                                   Note: 'MRI', 'PET', 'CT' and 'X-Ray' can also be used.
                                   This parameter is not case sensitive. Defaults to ``None``.
        :type allowed_modalities: ``list`` or ``tuple``
        :param save_dicom: if ``True``, save the DICOM images provided by The Cancer Imaging Archive 'as is'. Defaults to ``True``.
        :type save_dicom: ``bool``
        :param save_png: if ``True``, convert the DICOM images provided by The Cancer Imaging Archive to PNGs. Defaults to ``False``.
        :type save_png: ``bool``
        :param new_records_pull: if ``True``, download the data for the current search. If ``False``, use ``INSTANCE.records_db``.
        :type new_records_pull: ``bool``
        :return: a DataFrame with the record information.
        :rtype: ``Pandas DataFrame``
        """
        if new_records_pull:
            if not isinstance(self.current_query, pd.DataFrame):
                raise ValueError("`search()` must be called before `pull()`.")
            self._pull_time = datetime.now()
            self.records_db = self._records_db_gen(patient_limit=patient_limit,
                                                   collections_limit=collections_limit)
        elif not isinstance(self.records_db, pd.DataFrame):
            raise TypeError("`records_db` is not a DataFrame.")

        if save_png or save_dicom:
            self.records_db = self._Images.pull_images(records_db=self.records_db,
                                                       session_limit=session_limit,
                                                       save_png=save_png,
                                                       save_dicom=save_dicom,
                                                       allowed_modalities=allowed_modalities)

            # Add the new records_db datafame with the existing `cache_records_db`.
            self._tcia_cache_records_db_handler()

        return self.records_db



# Expose key Classes at the Subpackage level
from biovida.diagnostics.disease_ont_interface import DiseaseOntInterface
from biovida.diagnostics.disease_symptoms_interface import DiseaseSymptomsInterface



# coding: utf-8

"""

    Disease Ontology Interface
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import os
import re
import pickle
import requests
import numpy as np
import pandas as pd
from warnings import warn
from itertools import chain
from datetime import datetime

# Biovida support tools
from biovida.support_tools.support_tools import cln, header, items_null
from biovida.support_tools._cache_management import package_cache_creator


class DiseaseOntInterface(object):
    """

    Python Interface for Harvesting the `Disease Ontology <http://disease-ontology.org/>`_ Database.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
                   Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    def __init__(self, cache_path=None, verbose=True):
        self._verbose = verbose
        # Cache creation
        pcc = package_cache_creator(sub_dir='diagnostics',
                                    to_create=['disease_ontology'],
                                    cache_path=cache_path,
                                    verbose=verbose)
        self.root_path, self._created_disease_ont_dirs = pcc

        # The database itself
        self.disease_db = None
        self.db_date = None

    @staticmethod
    def _quote_value_parse(q):
        """

        Splits a string on the presence of quotes (") inside the string itself.
    
        :param q: a string containing quotes.
        :type q: ``str``
        :return: ``[text in quotes, text outside of quotes]``
        :rtype: ``list``
        """
        return list(map(cln, filter(None, q.split("\""))))
    
    def _def_url_parser(self, definition):
        """

        Separate the text of a disease definition from the list of reference URLs for that definition.
    
        :param definition: the raw text of the definition as represented in the Disease Ontology Database.
        :type definition: ``str``
        :return: a list of two ``tuple``s of the form:
                ``[('def', disease definition), ('def_urls', urls to def. sources)]``
        :rtype: ``list``
        """
        # Note: this method automatically replaces the underscores in the definition quotes.
        # This seems to be valid currently as this shortcut seems to invariably produce the same
        # result as would be obtained by following the list of mappings at the top of the .obo
        # database file. *This could change*.
        if definition.count("\"") != 2 or definition.count("[") != 1 or definition.count("]") != 1:
            return [("def", definition), ("def_urls", np.NaN)]

        # Separate the quote from the list of URLS
        parsed_definition = self._quote_value_parse(definition)
    
        # Extract the list of urls
        urls = parsed_definition[1].lower().replace("url:", "").replace("[", "").replace("]", "").split(", ")
    
        # Remove escape for the colon in the urls
        cleaned_urls = [u.replace("\:/", ":/") for u in urls]
    
        # Return the quote and the urls as separate entities
        return [("def", parsed_definition[0].replace("_", " ")), ("def_urls", cleaned_urls)]

    @staticmethod
    def _is_a_parser(is_a):
        """

        Separates the DOID from the text of the entry in an 'is_a' entry.

        :param is_a: a value corresponding to a term entry 'is_a' in the Disease Ontology database
        :type is_a: ``str``
        :return: list of tuples of the form:
                `` [("is_a", text of entry), ("is_a_doid", DOID number of entry)]``
        :rtype: ``list``
        """
        if " ! " not in is_a:
            return is_a
        parse_input = cln(is_a).split(" ! ")
        return [("is_a", parse_input[1]), ("is_a_doid", parse_input[0].upper().replace("DOID:", ""))]
    
    def _value_parser(self, k, v):
        """

        Parses the ``v`` evolved inside ``DiseaseOntInterface()._parsed_term_to_dict()`` based on ``k``.
    
        :param k: the k (key) passed inside of ``DiseaseOntInterface()._parsed_term_to_dict`` .
        :type k: ``str``
        :param v: the v (value) passed inside of ``DiseaseOntInterface()._parsed_term_to_dict``.
        :type v: ``str``
        :return: one of:

        - ``[('def', disease definition), ('def_urls', urls to def. sources)]``

        - ``[(KEY, value.replace("DOID:", ""))]``

        - ``[(KEY, information within quotes in v), (KEY_FLAG, list information in v)]``*

        - ``[(KEY, value)]``

        *Example: "'the quick brown fox' EXACT [info1, info2, info3]", where 'EXACT' is the FLAG.
        Here:

            - "information within quotes in v" = 'the quick brown fox'.

            - "list information in v": "[info1, info2, info3]" (string) --> ['info1', 'info2', 'info3']  (python list).

        :rtype: ``list``
        """
        if k == 'def':
            return self._def_url_parser(v)
        elif k == 'is_a':
            return self._is_a_parser(v)
        elif k in ('id', 'alt_id'):
            return [(k, v.upper().replace("DOID:", ""))]
        elif v.count("\"") == 2 and v.count("[") == 1 and v.count("]") == 1:
            # Split the true quote and the 'flags' (e.g., 'EXACT').
            parsed_v = self._quote_value_parse(v)
            # Split the flag and its corresponding list
            additional_v = re.split(r'\s(?=\[)', parsed_v[1])
            # Clean the flag list
            cleaned_flag = map(cln, additional_v[1].replace("[", "").replace("]", "").split(", "))
            # Filter the flag list
            related_info = list(filter(None, cleaned_flag))
            # Return the (key, quote) and the (key_flag, info).
            return [(k, parsed_v[0]), ("{0}_{1}".format(k, additional_v[0].lower()), related_info)]
        else:
            return [(k, v)]
    
    def _parsed_term_to_dict(self, parsed_term):
        """

        Parses an individual "[Term]" in the database and recasts it as a dictionary.
    
        :param parsed_term: a list of the form [[KEY, VALUE]...] as described
        :type parsed_term: ``list``
        :return: ``parsed_term`` converted into a dictionary where the information as well as a list of the keys 
                (future column names) which contain lists.
        :rtype: ``tuple``
        """
        d = dict()
        keys_with_lists = set()
        for (k, v) in parsed_term:
            parsed = self._value_parser(k, v=cln(v))
            for (kp, vp) in parsed:
                if kp not in d:
                    if isinstance(vp, list):
                        if len(vp):
                            keys_with_lists.add(kp)
                            d[kp] = vp
                        else:
                            d[kp] = np.NaN
                    else:
                        d[kp] = vp
                elif kp in d:
                    if isinstance(d[kp], list):
                        d[kp] += vp if isinstance(vp, list) else [vp]
                        keys_with_lists.add(kp)
                    elif items_null(d[kp]):
                        # In short, if the current value is NaN, replace with it with `vp` if
                        # and only if it is a list with nonzero length, otherwise leave as a NaN.
                        if isinstance(vp, list) and len(vp):
                            keys_with_lists.add(kp)
                            d[kp] = vp
                    else:
                        keys_with_lists.add(kp)
                        d[kp] = [d[kp], vp]
    
        return d, keys_with_lists
    
    def _do_term_parser(self, term):
        """
        
        This method splits a "[Term]" on line breaks ("\n"), the resultant list
        is then converted to a list of lists (where the sublists are invariably of length `2`),
        by splitting the elements into keys and values, e.g., ``['id', '000000']``.
        This list of lists is passed to ``DiseaseOntInterface()._parsed_term_to_dict()`` where it
        is converted into a dictionary.
        
        :param term: a single "[Term]" from the Disease Ontology Database.
        :type term: ``str``
        :return: a term which as been parsed by ``DiseaseOntInterface()._parsed_term_to_dict()``.
        :rtype: ``dict``
        """
        # Split the term on line breaks
        split_term = filter(None, cln(term).split("\n"))
    
        # Split each element in `term` on the ": " pattern.
        parsed_term = [i.split(": ", 1) for i in split_term]
    
        # Convert to a dict and return
        return self._parsed_term_to_dict(parsed_term)

    @staticmethod
    def _do_df_cleaner(data_frame, columns_with_lists):
        """

        This method cleans the final Disease Ontology database in the following ways:

        - converts columns which contains lists to strings.

        - lowers all strings in the following columns: 'name', 'synonym', 'subset' and 'is_a'.

        - converts the 'true' string in the 'is_obsolete' column to an actual python boolean ``True``.
    
        :param data_frame: the dataframe evolved in the ``DiseaseOntInterface()._harvest_engine()`` method.
        :type data_frame: ``Pandas DataFrame``
        :param columns_with_lists: a 'list' of columns in the dataframe which contain lists.
        :type columns_with_lists: ``iterable``
        :return: a dataframe with the above mentioned cleaning steps taken.
        :rtype: ``Pandas DataFrame``
        """
        # Homogenize columns with lists
        for c in columns_with_lists:
            data_frame[c] = data_frame[c].map(lambda x: "; ".join(x) if isinstance(x, list) else x,
                                              na_action='ignore')
    
        # Lower columns to make it easier to match in the future
        for c in ('name', 'synonym', 'subset', 'is_a'):
            data_frame[c] = data_frame[c].map(lambda x: str(x).lower(), na_action='ignore')
    
        # Convert 'true' in the 'is_obsolete' column to an actual python boolean ``True``.
        data_frame['is_obsolete'] = data_frame['is_obsolete'].map(
            lambda x: True if not items_null(x) and str(x).lower().strip() == 'true' else x,
            na_action='ignore')
    
        return data_frame

    def _extract_date_version(self, first_parsed_by_term):
        """

        Extracts the date the Disease Ontology data was created.

        :param first_parsed_by_term: the first element in the list obtained by splitting the database on '[Term]'.
        :type first_parsed_by_term: ``str``
        """
        try:
            extracted_date = re.search('data-version: (.*)\n', first_parsed_by_term).group(1)
            extracted_date_cleaned = "".join((i for i in extracted_date if i.isdigit() or i == "-")).strip()
            self.db_date = datetime.strptime(extracted_date_cleaned, "%Y-%m-%d")
        except:
            warn("\nCould not extract the date on which the Disease Ontology database was generated.")

    def _harvest_engine(self, disease_ontology_db_url, **kwargs):
        """

        This method Harvests and orchestrates and conversion of the Disease Ontology Database
        from '.obo' format to a Pandas DataFrame.

        :param disease_ontology_db_url: see: ``DiseaseOntInterface().pull()``.
        :type disease_ontology_db_url: ``str``
        """
        if 'obo_file' in kwargs:
            obo_file = kwargs.get('obo_file')
        else:
            # Open the file and discard [Typedef] information at the end of the file.
            obo_file = requests.get(disease_ontology_db_url, stream=True).text.split("[Typedef]")[0]
    
        # Parse the file by splitting on [Term].
        parsed_by_term = obo_file.split("[Term]\n")
    
        # Extract the date
        self._extract_date_version(parsed_by_term[0])

        # Convert to a list of dicts
        fully_parsed_terms = [self._do_term_parser(term) for term in parsed_by_term[1:]]
    
        # Extract the dicts
        list_of_dicts = [i[0] for i in fully_parsed_terms]
    
        # Extract keys (future column names) which contain lists
        keys_with_lists = filter(None, (i[1] for i in fully_parsed_terms))
    
        # Compress `keys_with_lists` to uniques.
        columns_with_lists = set(chain(*keys_with_lists))
    
        # Convert to a DataFrame, Clean and Return
        self.disease_db = self._do_df_cleaner(pd.DataFrame(list_of_dicts), columns_with_lists)

    def pull(self, download_override=False, disease_ontology_db_url='http://purl.obolibrary.org/obo/doid.obo'):
        """

        Pull (i.e., download) the Disease Ontology Database.

        Notes:

        - if a database is already cached, it will be used instead of downloading (use ``download_override`` to override).

        - multiple values are separated by semicolons followed by a space, i.e., "; ".

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :param disease_ontology_db_url: URL to the disease ontology database in '.obo' format.
                                        Defaults to 'http://purl.obolibrary.org/obo/doid.obo'.
        :type disease_ontology_db_url: ``str``
        :return: the Disease Ontology database as a DataFrame.
        :rtype: ``Pandas DataFrame``
        """
        save_path = os.path.join(self._created_disease_ont_dirs['disease_ontology'], "disease_ontology_db")
        db_path = "{0}.p".format(save_path)
        support_path = "{0}_support.p".format(save_path)

        if not os.path.isfile(db_path) or download_override:
            if self._verbose:
                header("Downloading Disease Ontology Database... ")
            self._harvest_engine(disease_ontology_db_url)
            self.disease_db.to_pickle(db_path)
            pickle.dump(self.db_date, open(support_path, "wb"))
        elif 'dataframe' not in str(type(self.disease_db)).lower():
            with open(support_path, "rb") as f:
                self.db_date = pickle.load(f)
            self.disease_db = pd.read_pickle(db_path)

        return self.disease_db



# coding: utf-8

"""

    General Tools for the Image Subpackage
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import numpy as np
from PIL import Image
from time import sleep
from scipy.misc import imread, imresize
from skimage.color.colorconv import rgb2gray

# General Support Tools
from biovida.support_tools.support_tools import tqdm


# ----------------------------------------------------------------------------------------------------------
# General Tools
# ----------------------------------------------------------------------------------------------------------


class NoResultsFound(Exception):
    pass


def sleep_with_noise(amount_of_time, mean=0.0, noise=0.75):
    """

    Sleep the current python instance by `amount_of_time`.

    :param amount_of_time: the amount of time to sleep the instance for.
    :type amount_of_time: ``int`` or ``float``
    :param mean: ``mean`` param of ``numpy.random.normal()``. Defaults to 0.0.
    :type mean: ``float``
    :param noise: ``scale`` param of ``numpy.random.normal()``. Defaults to 0.75.
    :type noise: ``float``
    """
    sleep(abs(amount_of_time + np.random.normal(loc=mean, scale=noise)))


def try_fuzzywuzzy_import():
    """

    Try to import the ``fuzzywuzzy`` library.

    """
    try:
        from fuzzywuzzy import process
        return process
    except ImportError:
        error_msg = "`fuzzy_threshold` requires the `fuzzywuzzy` library,\n" \
                    "which can be installed with `$ pip install fuzzywuzzy`.\n" \
                    "For best performance, it is also recommended that python-Levenshtein is installed.\n" \
                    "(`pip install python-levenshtein`)."
        raise ImportError(error_msg)


def load_image_rescale(path_to_image, gray_only=False):
    """

    Loads an image, converts it to grayscale and normalizes (/255.0).

    :param path_to_image: the address of the image.
    :type path_to_image: ``str``
    :param gray_only: if ``True``, simply convert to grayscale and return.
                      Otherwise, read-in, flatten and convert to grayscale. Default to ``False``.
    :type gray_only: ``bool``
    :return: the image as a matrix.
    :rtype: ``ndarray``
    """
    if gray_only:
        return rgb2gray(path_to_image) / 255.0
    else:
        return rgb2gray(imread(path_to_image, flatten=True)) / 255.0


def image_transposer(converted_image, image_size, axes=(2, 0, 1)):
    """

    Tool to resize and transpose an image (about given axes).

    :param converted_image: the image as a ndarray.
    :type converted_image: ``ndarray``
    :param image_size: to size to coerse the images to be, e.g., (150, 150)
    :type image_size: ``tuple``
    :param axes: the axes to transpose the image on.
    :type axes: ``tuple``
    :return: the resized and transposed image.
    :rtype: ``ndarray``
    """
    return np.transpose(imresize(converted_image, image_size), axes).astype('float32')


def load_and_scale_images(list_of_images, image_size, axes=(2, 0, 1), status=True, grayscale_first=False, desc=None):
    """

    Load and scale a list of images from a directory

    :param list_of_images: a list of paths to images.
    :type list_of_images: ``list`` or ``tuple``
    :param image_size: to size to coerce the images to be, e.g., (150, 150)
    :type image_size: ``tuple``
    :param axes: the axes to transpose the image on.
    :type axes: ``tuple``
    :param status: if ``True``, use `tqdm` to print progress as the load progresses.
    :type status: ``bool``
    :param grayscale_first: convert the image to grayscale first.
    :type grayscale_first: ``bool``
    :param desc: description to pass to ``tqdm``.
    :type desc: ``str`` or ``None``
    :return: the images as ndarrays nested inside of another ndarray.
    :rtype: ``ndarray``
    """
    # Source: https://blog.rescale.com/neural-networks-using-keras-on-rescale/
    def load_func(image):
        if 'ndarray' == type(image).__name__:
            converted_image = image
        else:
            # Load grayscale images by first converting them to RGB (otherwise, `imresize()` will break).
            if grayscale_first:
                loaded_image = Image.open(image).convert("LA")
                loaded_image = loaded_image.convert("RGB")
            else:
                loaded_image = Image.open(image).convert("RGB")
            converted_image = np.asarray(loaded_image)
        return image_transposer(converted_image, image_size, axes=axes)

    loaded = [load_func(image_name) for image_name in tqdm(list_of_images, desc=desc, disable=not status)]
    return np.array(loaded) / 255.0


def show_plt(image):
    """

    Use matplotlib to display an image (which is represented as a matrix).

    :param image: an image represented as a matrix.
    :type image: ``ndarray``
    """
    from matplotlib import pyplot as plt
    fig, ax = plt.subplots()
    ax.imshow(image, interpolation='nearest', cmap=plt.cm.gray)
    plt.show()



# Expose key tools at the Subpackage level
from biovida.support_tools.printing import dict_pprint
from biovida.support_tools.printing import pandas_pprint
from biovida.support_tools.utilities import train_val_test
from biovida.support_tools.utilities import reverse_train_val_test



# coding: utf-8

"""

    Border and Edge Detection
    ~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import numpy as np
import pandas as pd
from operator import sub
from functools import reduce


from biovida.images._image_tools import load_image_rescale


def _rounder(l, by=3):
    """

    Rounds all of the elements in an iterable.

    :param l: a list of values
    :type l: ``list`` or ``tuple``
    :param by:  If 0, the values will be converted to integers.
    :type by: ``int``
    :return: rounded numbers.
    :rtype: ``list``
    """
    t = int if by == 0 else float
    return list(map(lambda x: round(t(x), by), l))


def _deltas(iterable):
    """

    Compute the _deltas between all adjacent elements in ``iterable``.

    :param iterable: an iterable structure
    :type iterable: ``list`` or ``tuple``
    :return: an array with the delta between juxtaposed entries.
    :rtype: ``ndarray``
    """
    # Source: http://stackoverflow.com/a/2400875/4898004.
    return np.array([abs(j-i) for i, j in zip(iterable, iterable[1:])])


def _largest_n_values(arr, n):
    """

    Returns the index (i.e., rows or columns) of the largest n numbers

    :param arr: a numpy array
    :type arr: ``1D ndarray``
    :param n: the largest ``n`` values to return.
    :type n: ``int``
    :return: sorted list of the ``n`` largest values in ``arr`` (ascending order).
    :rtype: ``tuple``
    """
    return tuple(sorted(arr.argsort()[-n:]))


def _subsection(iterable, exclude, start, end):
    """

    Selects a subsection of an array, excluding a single subsection

    :param iterable: some iterable data structure, which can be indexed.
    :type iterable: ``list`` or ``tuple``
    :param exclude: the index to exclude.
    :type exclude: ``int``
    :param start: where to start the slice.
    :type start: ``int``
    :param end: where to end the slice
    :type end: ``int``
    :return: a list of the desired subsection.
    :rtype: ``list``
    """
    return iterable[start:exclude] + iterable[exclude+1:end]


def _anomaly_removal(numeric_array, window=2):
    """

    Replaces single numbers that do not match their homogeneous neighbourhood
    with the value of the homogeneous neighbors.

    E.g., 0, 0, 0, 0, 99, 0, 0, 0, 0 --> 0, 0, 0, 0, 0, 0, 0, 0, 0.

    :param numeric_array: must be a list or tuple. Numpy array will break this.
    :param window:
    :return: a list of smoothed values.
    :rtype: ``list``

    :Example:

    >>> row = [0.24677, 0.24677, 0.24677, 0.9, 0.24677, 0.24677, 0.24677]
    >>> for i, j in zip(row, _anomaly_removal(row, 3)):
    ...     print(i, "-->", j, ".Change Made: ", i != j)
        0.24677 --> 0.24677 .Change Made:  False
        0.24677 --> 0.24677 .Change Made:  False
        0.24677 --> 0.24677 .Change Made:  False
        0.9     --> 0.24677 .Change Made:  True
        0.24677 --> 0.24677 .Change Made:  False
        0.24677 --> 0.24677 .Change Made:  False
        0.24677 --> 0.24677 .Change Made:  False
    """
    # Note: the logic is somewhat complex here...a mistake is possible.

    if window <= 1 or window > len(numeric_array):
        raise ValueError("`window` must be greater than 1 and less than the length of `numeric_array`.")

    if window > 2:
        window += 1

    smoothed = list()
    for i in range(len(numeric_array)):
        # if can't look back
        if i < window:
            # If all forward neighbours are the same
            if len(set(numeric_array[i+1:i+window+1])) == 1:
                # Smooth using the next element
                smoothed.append(numeric_array[i+1])
            else:
                # if not, leave 'as is'
                smoothed.append(numeric_array[i])
        # if can look back
        if i >= window:
            if len(set(_subsection(numeric_array, i, i-window+1, i+window))) == 1:
                # smooth using the prior element (needed to prevent error at the end of the iterable).
                smoothed.append(numeric_array[i-1])
            else:
                smoothed.append(numeric_array[i])

    return smoothed


def _rolling_avg(iterable, window):
    """

    Computes the rolling average using ``Pandas``.

    :param iterable: some iterable data structure, which can be indexed.
    :type iterable: ``list`` or ``tuple``
    :param window: the window for the rolling average.
    :type window: ``int``
    :return: the rolling average of the iterable. Note: center = ``False``.
    :rtype: ``Pandas Series``
    """
    return pd.Series(iterable).rolling(center=False, window=window).mean().dropna()


def _array_cleaner(arr, round_by=5, anomaly_window=2):
    """

    Round array elements and remove anomalies using ``_anomaly_removal()``.

    :param arr: a numpy array
    :type arr: ``ndarray``
    :param round_by: how much to round the input array
    :type round_by: ``int``
    :param anomaly_window: window on which to check for a replace anomalies. See _anomaly_removal().
    :type anomaly_window: ``int``
    :return: an array of the smoothed values.
    :rtype: ``ndarray``
    """
    rounded_arr = _rounder(arr, round_by)
    return np.array(_anomaly_removal(rounded_arr, anomaly_window))


def _largest_n_changes_with_values(iterable, n):
    """

    Compute the index of the ``n`` largest _deltas the their associated values.

    :param iterable: an iterable data structure which supports indexing.
    :type iterable: ``list`` or ``tuple``
    :param n: ``n`` largest.
    :type n: ``int``
    :return: a list of the index with largest changes, along with the values themselves
    :rtype: ``list``
    """
    large_ds = _largest_n_values(_deltas(iterable), n)

    # Look around for true smallest value for each.
    # i.e., the index of the value which triggered the large delta.
    true_smallest = list()
    for d in large_ds:
        if d == 0 or len(iterable) == 2:
            neighbours = [0, 1]
        elif d == len(iterable):
            neighbours = [-1, 0]
        else:
            neighbours = [-1, 0, 1]
        options = [(d+i, iterable[d+i]) for i in neighbours]
        flr = [i for i in options if i[1] == min((j[1] for j in options))][0][0]
        true_smallest.append(flr)

    return [(d, iterable[d]) for d in true_smallest]


def _expectancy_violation(expected, actual, round_by=4):
    """

    Formula to compute the percent error.

    :param expected: the expected value for a given event.
    :type expected: ``float`` or ``int``
    :param actual: the actual value for a given event.
    :type actual: ``float`` or ``int``
    :param round_by: how much to round the result. Defaults to 4.
    :type round_by: ``int``
    :return: percent error
    :type: ``float``
    """
    # ToDo: explore better solutions.
    # This just blocks division by zero (given all values will be >= 0)
    expected += 0.0001
    actual += 0.0001

    val = round(float(abs(expected - actual) / expected), round_by)
    return round(float(val), round_by)


def _largest_median_inflection(averaged_axis_values, axis, n_largest_override=None):
    """

    Finds the values in a vector which diverge most strongly from the median value.

    :param averaged_axis_values: a 1D array which has been computed by averaging about a given axis in a matrix.
    :type averaged_axis_values: ``ndarray``
    :param axis: 0 for columns; 1 for rows.
    :type axis: ``int``
    :return: a list of rows/columns which diverge from the median,
             and the percent error as a metric of the strength of this divergence.
    :rtype: ``list``
    """
    if isinstance(n_largest_override, int):
        n_largest = n_largest_override
    elif axis == 1:
        n_largest = 4
    else:
        n_largest = 2

    # Position of largest changes
    large_inflections = _largest_n_changes_with_values(averaged_axis_values, n_largest)

    # Sort by position
    large_inflections_sorted = sorted(large_inflections, key=lambda x: x[0])

    # Compute the median for the whole axis
    median = np.median(averaged_axis_values)

    # Compute the how much the signal deviated from the median value (0-1).
    median_deltas = [(i, _expectancy_violation(median, j)) for (i, j) in large_inflections_sorted]

    if isinstance(n_largest_override, int):  # neighborhood search in _largest_n_changes_with_values may --> duplicates.
        return median_deltas
    elif axis == 1:
        return median_deltas[1:]
    elif axis == 0:
        return median_deltas


def _zero_var_axis_elements_remove(image, axis, rounding=3):
    """

    Replaces, by axis, matrix elements with approx. no variance
    (technically using the standard deviation here).

    :param image: an image represented as an array.
    :type image: ``ndarray``
    :param axis: 0 for columns; 1 for rows.
    :type axis: ``int``
    :param rounding: how much to round the standard deviation values for a given row/column.
    :type rounding: ``int``
    :return: a matrix with rows/columns with standard deviation == 0 replaced with zero vectors.
    :rtype: ``ndarray``
    """
    zero_var_items = np.where(np.round(np.std(image, axis=axis), rounding) == 0)
    if axis == 0:
        image[:, zero_var_items] = [0]
    elif axis == 1:
        image[zero_var_items] = [0]
    return image


def edge_detection(image, axis=0, n_largest_override=None):
    """

    Detects edges within an image.

    :param image: an image represented as a matrix
    :type image: ``ndarray``
    :param axis: 0 for columns; 1 for rows.
    :type axis: ``int``
    :param n_largest_override: override the defaults for the number of inflections to report
                               when searching the image along a given axis.
    :type n_largest_override: ``int`` or ``None``
    :return: the location of large inflections (changes) in the image along a given axis.
    :rtype: ``list``
    """
    # Set rows with no ~variance to zero vectors to eliminate their muffling effect on the signal.
    image = _zero_var_axis_elements_remove(image, axis)

    # Average the remaining values
    # ToDo: it's not not clear if ``_array_cleaner()`` helps much...after all, the vector has already been averaged.
    averaged_axis_values = _array_cleaner(np.mean(image, axis=axis))
    return _largest_median_inflection(averaged_axis_values, axis, n_largest_override)


def _weigh_evidence(candidates, axis_size, signal_strength_threshold, min_border_separation, buffer_multiplier=1/15):
    """

    Weigh the evidence that a true border has been detected.

    :param candidates: a list of lists or list of tuples
    :type candidates: ``list``
    :param axis_size: how long a given axis is, e.g., 512 (for a 512x256 image).
    :type axis_size: ``int``
    :param signal_strength_threshold: how strong the inflection must be relative
                                      to the median for the image. Must be between 0 and 1.
    :type signal_strength_threshold: ``float``
    :param min_border_separation: a value between 0 and 1 that determines the proportion of the axis
                                  that two edges must be separated for them to be considered borders.
                                  (i.e., ``axis_size`` * ``min_border_separation``)
    :type min_border_separation: ``float``
    :param buffer_multiplier: How far from the midpoint two lines must to be considered a border.
                              (calculation: midpoint +/- (axis_size * buffer_multiplier). Defaults to 1/15.
    :type buffer_multiplier: ``int`` or ``float
    :return: None if it was 'decided' that there was no enough evidence, else the candidates are returned 'as is'.
    :rtype: ``list`` or ``None``
    """
    midpoint = (axis_size / 2)
    lu_buffer = midpoint - (axis_size * buffer_multiplier)  # left/upper
    rl_buffer = midpoint + (axis_size * buffer_multiplier)  # right/lower

    conclusion = None
    if all(x[1] >= signal_strength_threshold for x in candidates):
        if abs(reduce(sub, [i[0] for i in candidates])) >= (min_border_separation * axis_size):
            if candidates[0][0] < lu_buffer and candidates[1][0] > rl_buffer:
                conclusion = candidates

    return conclusion


def _lower_bar_detection(image_array, lower_bar_search_space, signal_strength_threshold, cfloor=None):
    """

    Executes a single pass looking for a lower bar.

    :param image_array: an image represented as a numpy array.
    :type image_array: ``ndarray``
    :param lower_bar_search_space: a value between 0 and 1 specify the proportion of the image
                                   to search for a lower bar (e.g., 0.90).
    :type lower_bar_search_space: ``float``
    :param signal_strength_threshold: a value between 0 and 1 specify the signal strength required
                                      for an area required to be considered a 'lower bar'.
                                      Internally, this is measured as a location deviation from
                                      the median signal strength of the average image.
    :type signal_strength_threshold: ``int``
    :param cfloor: check floor. If None, the floor is the last image in the photo.
    :type cfloor: ``int`` or ``None
    :return: the height (y position) of the top of the lower bar (given one pass of this algorithm).
    :rtype: ``int``
    """
    # Compute the location to crop the image
    cut_off = int(image_array.shape[0] * lower_bar_search_space)
    flr = image_array.shape[0] if not isinstance(cfloor, int) else cfloor

    if abs(cut_off - flr) < 3:
        return cfloor if cfloor is not None else None  # redundant

    lower_image_array = image_array.copy()[cut_off:flr]

    lower_bar_candidates = edge_detection(image=lower_image_array, axis=1, n_largest_override=8)
    thresholded_values = list(set([i[0] + cut_off for i in lower_bar_candidates if i[1] > signal_strength_threshold]))

    if not len(thresholded_values):
        return None

    # Return the averaged guess
    mean_thresholded_values = np.mean(thresholded_values).astype(int)
    return int(mean_thresholded_values)


def lower_bar_analysis(image_array, lower_bar_search_space, signal_strength_threshold, lower_bar_second_pass):
    """

    Executes a two passes looking for a lower bar.
    This can be helpful if the lower border is striated in such a way that could confuse the
    algorithm. Namely, two the space between two lines of text can confuse the algorithm.
    The first pass with truncate the first line; the second pass will truncate the line above it.

    :param image_array: an image represented as a matrix.
    :type image_array: ``2D ndarray``
    :param lower_bar_search_space: a value between 0 and 1 specify the proportion of the image
                                   to search for a lower bar (e.g., 0.90).
    :type lower_bar_search_space: ``float``
    :param signal_strength_threshold: a value between 0 and 1 specify the signal strength required
                                      for an area required to be considered a 'lower bar'.
                                      This is measured as a absolute value of the difference between
                                      a location and the median signal strength of the average image.
    :type signal_strength_threshold: ``int`` or ``float``
    :param lower_bar_second_pass: if ``True`` perform a second pass on the lower bar.
    :type lower_bar_second_pass: ``bool``
    :return: the location of the start of the lower bar (i.e., edge).
    :rtype: ``int``

    :Example:

    Take the following lower border in an image:

    .. code-block:: python

             --- Image Here ---
        _____________________________
        The quick brown fox jumped
        over jumps over the lazy dog
        -----------------------------

    One pass of this procedure should produce:

    .. code-block:: python

             --- Image Here ---
        _____________________________
        The quick brown fox jumped
        -----------------------------

    Following this up with a second pass should yeild:

    .. code-block:: python

             --- Image Here ---
        _____________________________

    """
    first_pass_rslt = _lower_bar_detection(image_array, lower_bar_search_space, signal_strength_threshold)
    
    if lower_bar_second_pass:
        second_pass_rslt = _lower_bar_detection(image_array,
                                                lower_bar_search_space,
                                                signal_strength_threshold,
                                                cfloor=first_pass_rslt)
    else:
        second_pass_rslt = None

    if second_pass_rslt is None:
        return first_pass_rslt
    else:
        return second_pass_rslt


def border_detection(image,
                     signal_strength_threshold=0.25,
                     min_border_separation=0.15,
                     lower_bar_search_space=0.9,
                     report_signal_strength=False,
                     rescale_input_ndarray=True,
                     lower_bar_second_pass=True):
    """

    Detects the borders and lower bar in an image.

    At a high level, this algorithm works as follows:
       1. Along a given axis (rows or columns), vectors
          with standard deviation approximately equal to 0 are replaced with zero vectors. (a).
       2. Values are averaged along this same axis.
          This produces a signal (which can be visualized as a line graph). (b).
       3. The median value for this signal is computed. (c).
       4. The ``n`` points for which are the furthest, in absolute value, from the median are selected.
       5. The signal strength of the ``n`` points is quantified using percent error, where the median value
          is used as the expected value.
       6. Candidates for border pairs (e.g., left and right borders) are then weighed based on three lines of evidence.
          Namely, their signal strength, how separated they are and their absolute distance
          from the image's midline (about the corresponding axis). If a candidate fails to meet any of these criteria,
          it is rejected.
       7. The evidence for a lower bar concerns only its signal strength, though only an area of image below a given
          height is considered when trying to locate it. A double pass, the default, will try a second time to find
          another lower bar (for reasons explained in the docstring for the ``lower_bar_analysis()`` function).
          Regardless of whether or not the second pass could find a second edge, all of the edges detect are averaged
          and returned as an integer. If no plausible borders could be found, ``None`` is returned.

    (a) This reduces the muffling effect that areas with solid color can have on step 2.
    (b) Large inflections after areas with little change suggest a transition from a solid background to an image.
    (c) The median is used here, as opposed to the average, because it is more robust against outliers.

    :param image: a path to an image or an image represented as a 2D ndarray.

                 .. warning::

                        If a ``ndarray`` is passed, it should be the output of the
                        ``biovida.images.image_tools.load_image_rescale()`` function.
                        Without this preprocessing, this function's stability is not assured.

    :type image: ``str`` or ``2D ndarray``
    :param signal_strength_threshold: a value between 0 and 1 specify the signal strength required
                                      for an area required to be considered a 'lower bar'.
                                      Internally, this is measured as a location deviation from
                                      the median signal strength of the average image.
    :type signal_strength_threshold: ``float``
    :param min_border_separation: a value between 0 and 1 that determines the proportion of the axis
                                  that two edges must be separated for them to be considered borders.
                                  (i.e., ``axis_size`` * ``min_border_separation``)
    :type min_border_separation: ``float``
    :param lower_bar_search_space: a value between 0 and 1 specify the proportion of the image
                                   to search for a lower bar (e.g., 0.90). Set to ``None`` to disable.
    :type lower_bar_search_space: ``float``
    :param report_signal_strength: if ``True`` include the strength of the signal suggesting the existence of an edge.
                                   Defaults to ``False``.
    :type report_signal_strength: ``bool``
    :param rescale_input_ndarray: if True, rescale a ``2D ndarray`` passed to ``image``.
    :param lower_bar_second_pass:  if ``True`` perform a second pass on the lower bar. Defaults to ``True``.
    :type lower_bar_second_pass: ``bool``
    :type rescale_input_ndarray: ``bool``
    :return: a dictionary of the form:
             ``{'vborder': (left, right) or None, 'hborder': (upper, lower) or None, 'hbar': int or None}``

            - 'vborder' gives the locations of vertical borders.

            - 'hborder' gives the locations of horizontal borders.

            - 'hbar' gives the location for the top of the horizontal bar at the bottom of the image.

    :rtype: ``dict``
    """
    if 'numpy' in str(type(image)):
        image_copy = image.copy()
        if rescale_input_ndarray:
            image_array = load_image_rescale(image_copy, gray_only=True)
        else:
            image_array = image_copy
    elif isinstance(image, str):
        image_array = load_image_rescale(image)
    else:
        raise ValueError("`image_array` must be either a path to an image or the image as a 2D ndarray.")

    d = dict.fromkeys(['vborder', 'hborder', 'hbar'], None)

    v_edge_candidates = edge_detection(image_array, axis=0)
    d['vborder'] = _weigh_evidence(v_edge_candidates,
                                   image_array.shape[1],      # Note: we have (rows, columns) = (height, width).
                                   signal_strength_threshold,
                                   min_border_separation)

    h_border_candidates = edge_detection(image_array, axis=1)

    # Run Analysis. This excludes final element in `h_border_candidates` as including a third element
    # is simply meant to deflect the pull of the lower bar, if present.
    d['hborder'] = _weigh_evidence(h_border_candidates[:2],
                                   image_array.shape[0],
                                   signal_strength_threshold,
                                   min_border_separation)

    d['hbar'] = lower_bar_analysis(image_array=image_array,
                                   lower_bar_search_space=lower_bar_search_space,
                                   signal_strength_threshold=signal_strength_threshold,
                                   lower_bar_second_pass=lower_bar_second_pass)

    if report_signal_strength:  # ToDo: not working for 'hbar'.
        return d
    else:
        return {k: tuple([i[0] for i in v]) if isinstance(v, list) else (v[0] if isinstance(v, (list, tuple)) else v)
                for k, v in d.items()}


def _lines_plotter(path_to_image):
    """

    Visualizes the default ``border_detection()`` settings.

    :param path_to_image: path to the image
    :type path_to_image: ``str``
    """
    from matplotlib import pyplot as plt
    from matplotlib import collections as mc

    image = load_image_rescale(path_to_image)
    analysis = {k: v for k, v in border_detection(image=image).items() if v is not None}

    h, w = image.shape
    h_lines_explicit = [[(0, i), (w, i)] for i in analysis.get('hborder', [])]
    v_line_explicit = [[(i, 0), (i, h)] for i in analysis.get('vborder', [])]
    lines = h_lines_explicit + v_line_explicit

    if "int" in str(type(analysis.get('hbar', None))):
        lines += [[(0, int(analysis['hbar'])), (w, int(analysis['hbar']))]]

    if len(lines):
        line_c = mc.LineCollection(lines, colors=['r'] * len(lines), linewidths=2)
        fig, ax = plt.subplots()
        ax.add_collection(line_c)
        ax.imshow(image, interpolation='nearest', cmap=plt.cm.gray)
        plt.show()
        print(analysis)
        return True
    else:
        fig, ax = plt.subplots()
        ax.text(0, 0, "No Results to Display", fontsize=15)
        ax.imshow(image, interpolation='nearest', cmap=plt.cm.gray)
        # print("No Results to Display.")
        print(analysis)
        return False



# coding: utf-8

"""

    Utilities
    ~~~~~~~~~

"""
import os
import shutil
import numpy as np
from PIL import Image
from warnings import warn
from scipy.ndimage import imread
from os.path import join as os_join
from os.path import basename as os_basename

# General Support Tools
from biovida.support_tools.support_tools import (tqdm,
                                                 dicom,
                                                 is_int,
                                                 isclose,
                                                 is_numeric,
                                                 list_to_bulletpoints,
                                                 InsufficientNumberOfFiles,
                                                 directory_existence_handler)


_TVT_GROUPS = ('train', 'validation', 'test')


# ----------------------------------------------------------------------------------------------------------
# Splitting Files into Train, Validation and Test
# ----------------------------------------------------------------------------------------------------------


def _subdirectories_in_path(path, to_block):
    """
    
    Generate a list of subdirectories in ``path``, excluding ``to_block``.

    :param path: a system path.
    :type path: ``str``
    :param to_block: a list of directories to exclude from the output.
    :type to_block: ``iterable``
    :return: see description.
    :rtype: ``list``
    """
    return [os_join(path, i) for i in os.listdir(path) if os.path.isdir(os_join(path, i)) and i not in to_block]


def _train_val_test_error_checking(target_dir, action, group_files_dict, tvt, allowed_actions):
    """

    Check for possible errors for before spinning up engine.

    :param target_dir: see ``train_val_test()``.
    :type target_dir: ``str``
    :param action: see ``train_val_test()``.
    :type action: ``str``
    :param group_files_dict: as evolved in side ``train_val_test()``.
    :type group_files_dict: ``dict``
    :param tvt: as evolved in side ``train_val_test()``.
    :type tvt: ``dict``
    """
    if not tvt:
        raise ValueError("At least one of `train`, `validation` and `test` must be numeric.")
    for k, v in tvt.items():
        if v is True:
            raise ValueError("`{0}` cannot be `True`".format(k))
    if not isclose(sum(tvt.values()), 1):
        raise ValueError("The following parameters do not sum to 1:\n"
                         "{0}.".format(", ".join(sorted(tvt.keys()))))
    # if isinstance(data, dict) and not all(isinstance(i, (list, tuple)) for i in data.values()):
    #     raise TypeError("The values of `data` must be lists or tuples.")
    if action not in allowed_actions:
        raise ValueError("`action` must be one of:\n"
                         "{0}.".format(list_to_bulletpoints(allowed_actions)))

    min_number_of_files = len(tvt.keys()) * len(group_files_dict.keys())
    for k, v in group_files_dict.items():
        if len(v) < min_number_of_files:
            raise InsufficientNumberOfFiles("\nThe '{0}' subdirectory/subgroup only contains {1} files,\n"
                                            "which is too few to distribute over {2} target locations.\n"
                                            "Calculation: len([{3}]) * len([{4}]) = {2}.".format(
                                            k, len(v), min_number_of_files, ", ".join(tvt.keys()),
                                            ", ".join(group_files_dict.keys())))


def _existing_files_dict_gen(directory, to_block):
    """

    Generate a dictionary of files in ``directory``.
    
    :param directory: a system path.
    :type directory: ``str``
    :param to_block: a list of subdirectory names to ignore.
    :type to_block: ``iterable``
    :return: a dictionary of the form: ``{subdirectory in ``directory``: [file_path, file_path, ...], ...}``.
    :rtype: ``dict``
    """
    directories = [i for i in os.listdir(directory) if os.path.isdir(os_join(directory, i))]

    def list_files(path):
        return [os_join(path, i) for i in os.listdir(path) if not i.startswith('.')]

    return {d: list_files(os_join(directory, d)) for d in directories if d not in to_block}


def _train_val_test_dict_sort(tvt):
    """

    Sort a dictionary with any of 'train', 'validation', 'test'
    as keys into that order.

    :param tvt: a dictionary containing 'train', 'validation', 'test' keys.
    :type tvt: ``dict``
    :return: ``tvt`` as a list of tuples
    :rtype: ``list``
    :raises ``KeyError``: if any key other than 'train', 'validation' or 'test' is in ``tvt``
    """
    order_dict = {'train': 1, 'validation': 2, 'test': 3}
    return sorted(tvt.items(), key=lambda x: order_dict.get(x[0]))


def _list_divide(l, tvt, random_state):
    """

    Splits a list, ``l``, into proportions described in ``tvt``.

    :param l: a 'list' to be split according to ``tvt``.
    :type l: ``list``
    :param tvt: as evolved in side ``train_val_test()``.
    :type tvt: ``dict``
    :param random_state: set a seed for random shuffling. Similar to ``sklearn.model_selection.train_test_split``.
    :type random_state: ``int``
    :return: a dictionary of the form: ``{tvt_key_1: [file_path, file_path, ...], ...}``. Sorted to ensure
             generation order is train --> validation --> test.
    :rtype: ``list``

    .. note::

        This function will randomly shuffle ``l``.

    :Example:

    >>> l = ['file/path/image_1.png', 'file/path/image_2.png', 'file/path/image_3.png', 'file/path/image_4.png']
    >>> tvt = {'validation': 0.5, 'train': 0.5}
    >>> _list_divide(l, tvt, random_state=None)
    ...
    [('train', ['file/path/image_4.png', 'file/path/image_2.png']),
    ('validation', ['file/path/image_1.png', 'file/path/image_3.png'])]
    
    .. note::
    
        This function also works with lists of ndarrays, e.g.,
        ``l = [ndarray, ndarray, ndarray, ...]``.

    """
    if is_int(random_state):
        np.random.seed(random_state)

    l_shuffled = np.random.permutation(l).tolist()
    tvt_sorted = _train_val_test_dict_sort(tvt)

    left, divided_dict = 0, dict()
    for e, (k, v) in enumerate(tvt_sorted, start=1):
        right = len(l_shuffled) * v
        right_full = int(left) + int(right) if e != len(tvt.keys()) else len(l_shuffled)
        divided_dict[k] = l_shuffled[int(left):right_full]
        left += right
    return _train_val_test_dict_sort(divided_dict)


def _file_paths_dict_to_ndarrays(dictionary, dimensions, stack, verbose=True):
    """

    Converts values (list of file paths) of a ``dictionary`` to ``ndarray``s.

    :param dictionary: a nested dictionary, where values for the inner nest are iterables of file paths.
    :type dictionary: ``dict``
    :param dimensions: dimensions of ``dictionary``. ``1`` for an un-nested dictionary, ``2`` for a nested dictionary.
    :type dimensions: ``int``
    :param stack: if ``True``, stack 3D volumes and time-series images.
    :type stack: ``bool``
    :param verbose: if ``True``, print additional information and a conversion status bar.
    :type verbose: ``bool``
    :return: the values for the inner nest as replaced with ``ndarrays``.
    :rtype: ``dict``
    """
    def identity_func(x):
        return x

    if verbose:
        def desc_tqdm(x):
            return tqdm(x, desc='Generating ndarrays')
        if dimensions == 1:
            status_inner, status_outer = desc_tqdm, identity_func
        elif dimensions == 2:
            status_inner, status_outer = identity_func, desc_tqdm
    else:
        status_inner = status_outer = identity_func

    def is_dicom(path):
        return any(path.lower().endswith(d) for d in ('.dicom', '.dcm', '.dcm30'))

    def load_dicom(path):
        try:
            return dicom.read_file(path).pixel_array
        except:
            warn("\nCould not load the following "
                 "image as an ndarray:\n{0}".format(path))
            return None

    def drop_none(l):
        return [i for i in l if i is not None]

    def load_as_ndarrays(paths):
        n_dicom = [p for p in paths if is_dicom(p)]
        if not len(n_dicom):
            return np.array([imread(p) for p in paths])
        elif len(n_dicom) == len(paths):
            dicoms = drop_none([load_dicom(p) for p in paths])
            if stack and len(dicoms) > 1 and len({d.shape for d in dicoms}) == 1:
                return np.stack(dicoms)
            else:
                return np.array(dicoms)
        else:
            raise TypeError("Mixture of DICOM and non-DICOM files provided.")

    def values_to_ndarrays(d):
        return {k2: load_as_ndarrays(v2) for k2, v2 in status_inner(d.items())}

    if dimensions == 1:
        return values_to_ndarrays(dictionary)
    elif dimensions == 2:
        return {k: values_to_ndarrays(v) for k, v in status_outer(dictionary.items())}


def _print_output_dict_structure(output_dict):
    """Print the structure of the `output_dict`
    generated by `_train_val_test_engine()`."""
    print("\nStructure:\n")
    for k, v in _train_val_test_dict_sort(output_dict):
        print("- '{0}':\n{1}".format(k, list_to_bulletpoints(v)))


def _train_val_test_engine(action, tvt, group_files_dict, target_path, random_state, allowed_actions, verbose):
    """

    Engine to power ``train_val_test()``.

    :param action: 'copy', 'move' or 'write_ndarray'
    :type action: ``str`` or ``None``
    :param tvt: as evolved in side ``train_val_test()``.
    :type tvt: ``dict``
    :param group_files_dict: as evolved in side ``train_val_test()`` -- ``{'group': [file_path, file_path, ...], ...}``.
                             
                             
                    .. note::
                    
                        If ``action='write_ndarray'``, this this data is expected to be of the form:
                        `{'group': [(ndarray, PATH), (ndarray, PATH), ...], ...}``.
    
    :type group_files_dict: ``dict``
    :param target_path: see ``train_val_test()``.
    :type target_path: ``str`` or ``None``
    :param random_state: set a seed for random shuffling. Similar to ``sklearn.model_selection.train_test_split``.
    :type random_state: ``None`` or ``int``
    :param allowed_actions: a list of allowed values for ``action``.
    :type allowed_actions: ``list`` or ``tuple``
    :param verbose: see ``train_val_test()``.
    :type verbose: ``bool``
    :return: a nested dictionary of the form ``{'train'/'val'/'test': {group_files_dict.key: [file, file, ...], ...}, ...}``.
    :rtype: ``dict``
    """
    _train_val_test_error_checking(target_dir=target_path, action=action,
                                   group_files_dict=group_files_dict, tvt=tvt,
                                   allowed_actions=allowed_actions)

    def output_dic_value(value2):
        return [path for (nd, path) in value2] if action == 'write_ndarray' else value2

    if action == 'copy':
        shutil_func = shutil.copy2
    elif action == 'move':
        shutil_func = shutil.move

    output_dict = dict()
    for k, v in tqdm(group_files_dict.items(), desc='Divvying Groups', disable=not verbose):
        for k2, v2 in _list_divide(v, tvt, random_state=random_state):
            out_value = output_dic_value(value2=v2)
            if k2 not in output_dict:
                output_dict[k2] = {k: out_value}
            else:
                output_dict[k2][k] = out_value
            if action in ('copy', 'move', 'write_ndarray'):
                desc = 'Processing {0} ({1})'.format(os_basename(k), k2)
                target = directory_existence_handler(os_join(target_path, os_join(k2, k)), allow_creation=True)
                if action in ('copy', 'move'):
                    for i in tqdm(v2, desc=desc, leave=False, disable=not verbose):
                        shutil_func(i, os_join(target, os_basename(i)))
                elif action == 'write_ndarray':
                    for (nd, path) in tqdm(v2, desc=desc, leave=False, disable=not verbose):
                        Image.fromarray(nd).save(os_join(target, os_basename(path)))

    if verbose:
        _print_output_dict_structure(output_dict)

    return output_dict


def _tvt_dict_gen(d):
    """Extract those of the train, validation, test (tvt)
     params which are numeric."""
    return {k: v for k, v in d.items() if k in _TVT_GROUPS and is_numeric(v)}


def train_val_test(data,
                   train,
                   validation,
                   test,
                   target_dir=None,
                   action='copy',  # ToDo: change to 'ndarray'.
                   delete_source=False,
                   stack=True,
                   random_state=None,
                   verbose=True):
    """

    Split ``data`` into any combination of the following: ``train``, ``validation`` and/or ``test``.

    :param data: 
    
        - a dictionary of the form: ``{group_name: [file_path, file_path, ...], ...}``.
        - the directory containing the data. This directory should contain subdirectories (the categories)
          populated with the files.

        .. warning::

                If a directory is passed, subdirectories therein entitled 'train', 'validation' and 'test'
                will be ignored.

    :type data: ``dict`` or ``str``
    :param train: the proportion images in ``data`` to allocate to ``train``. If ``False`` or ``None``,
                  no images will be allocated.
    :type train: ``int``, ``float``, ``bool`` or ``None``
    :param validation: the proportion images in ``data`` to allocate to ``validation``. If ``False``
                       or ``None``, no images will be allocated.
    :type validation: ``int``, ``float``, ``bool`` or ``None``
    :param test: the proportion images in ``data`` to allocate to ``test``. If ``False`` or ``None``,
                 no images will be allocated.
    :type test: ``int``, ``float``, ``bool`` or ``None``
    :param target_dir: the location to output the images to (if ``action=True``). If ``None``, the output location will
                       be ``data``. Defaults to ``None``.
    :type target_dir: ``str`` or ``None``
    :param action: one of: 'copy', 'ndarray'.

        - if ``'copy'``: copy from files from ``data`` to ``target_dir`` (default).

        - if ``'move'``: move from files from ``data`` to ``target_dir``.

        - if ``'ndarray'``: return a nested dictionary of ``ndarray`` ('numpy') arrays.

        .. warning::

                **Using 'move' directly on files in a cache is not recommended**.
                However, if this action is performed, the corresponding class must be
                reinstated to allow the ``cache_records_db`` database to update.

    :type action: ``str``
    :param delete_source: if ``True`` delete the source subdirectories in ``data`` after copying is complete. Defaults to ``False``.

                          .. note::

                                This can be useful for transforming a directory 'in-place',
                                e.g., if ``data`` and ``target_dir`` are the same and ``delete_source=True``.

    :type delete_source: ``bool``
    :param stack: if ``True``, stack 3D volumes and time-series images when ``action='ndarray'``. Defaults to ``True``.
    :type stack: ``bool``
    :param random_state: set a seed for random shuffling. Similar to ``sklearn.model_selection.train_test_split``.
                         Defaults to ``None``.
    :type random_state: ``None`` or ``int``
    :param verbose: if ``True``, print the resultant structure. Defaults to ``True``.
    :type verbose: ``bool``
    :return:

        a dictionary of the form: ``{one of 'train', 'validation' or 'test': {subdirectory in `data`: [file_path, file_path, ...], ...}, ...}``.

        - if ``action='copy'``: the dictionary returned will be exactly as shown above.

        - if ``action='ndarray'``: 'file_path' will be replaced with the image as a ``ndarray`` and the list will be
          ``ndarray``s, i.e., ``array([Image Matrix, Image Matrix, ...])``.

    :rtype: ``dict``
    :raises ``ValueError``: if the combination of ``train``, ``validation``, ``test`` which which were passed
                            numeric values (i.e., ``int`` or ``float``) do not sum to 1.


    .. note::

          Files are randomly shuffled prior to assignment.

    .. warning::

           In the case of division with a remainder, preference is as follows: `train` < `validation` < `test`.
           For instance, if we have ``train=0.7``, ``validation=0.3``, and the number of files in a given subdirectory
           equal to ``6`` (as in the final example below) the number of files allocated to `train` will be rounded
           down, in favor of `validation` obtaining the final file. In this instance, `train` would obtain
           `4` files (``floor(0.7 * 6) = 4``) and `validation` would obtain 2 files (``ceil(0.3 * 6) = 2``).


    :Example:

    The examples below use a sample directory entitled ``images``.

    This is its structure:

    .. code-block:: bash

        $ tree /path/to/data/images
        ├── ct
        │   ├── ct_1.png
        │   ├── ct_2.png
        │   ├── ct_3.png
        │   ├── ct_4.png
        │   ├── ct_5.png
        │   └── ct_6.png
        └── mri
            ├── mri_1.png
            ├── mri_2.png
            ├── mri_3.png
            ├── mri_4.png
            ├── mri_5.png
            └── mri_6.png

    |
    | **Usage 1**: Obtaining ndarrays

    >>> from biovida.support_tools import train_val_test
    >>> tt = train_val_test(data='/path/to/data/images', train=0.7, validation=None, test=0.3,
    ...                     action='ndarray')

    The resultant ndarrays can be unpacked into objects as follows:

    >>> train_ct, train_mri = tt['train']['ct'], tt['train']['mri']
    >>> test_ct, test_mri = tt['test']['ct'], tt['test']['mri']

    |
    | **Usage 2**: Reorganize a Directory In-place

    >>> from biovida.support_tools import train_val_test
    >>> tv = train_val_test(data='/path/to/data/images', train=0.7, validation=0.3, test=None,
    ...                     action='copy', delete_source=True)

    Which results in the following structure:

    .. code-block:: bash

        $ tree /path/to/data/images
        ├── train
        │   ├── ct
        │   │   ├── ct_4.png
        │   │   ├── ct_5.png
        │   │   ├── ct_3.png
        │   │   └── ct_1.png
        │   └── mri
        │       ├── mri_2.png
        │       ├── mri_1.png
        │       ├── mri_5.png
        │       └── mri_4.png
        └── validation
            ├── ct
            │   ├── ct_2.png
            │   └── ct_6.png
            └── mri
                ├── mri_3.png
                └── mri_6.png

    .. note::

        The following changes to **Usage 2** would preserve the original ``ct`` and ``mri`` directories:

        - setting ``delete_source=False`` (the default) and/or
        - providing a path to ``target_dir``, e.g., ``target_dir='/path/to/output/output_data'``

    """
    if action in ('copy', 'move'):
        if isinstance(target_dir, str):
            target_path = target_dir
        elif isinstance(data, str) and not isinstance(target_dir, str):
            target_path = data
            warn("`target_path` is not a string; using `data` as the output location")
        else:
            raise TypeError("`target_dir` must be a system path if `data` is not.")
    else:
        target_path = None

    existing_dirs = _subdirectories_in_path(data, to_block=_TVT_GROUPS) if isinstance(data, str) else None
    tvt = _tvt_dict_gen(locals())

    if not isinstance(delete_source, bool):
        raise TypeError("`delete_source` must be a boolean.")

    if isinstance(data, str):
        group_files_dict = _existing_files_dict_gen(directory=data, to_block=_TVT_GROUPS)
    elif isinstance(data, dict):
        group_files_dict = data
    else:
        raise TypeError("`data` must be a string or dictionary.")

    output_dict = _train_val_test_engine(action=action, tvt=tvt, group_files_dict=group_files_dict,
                                         target_path=target_path, random_state=random_state,
                                         allowed_actions=('copy', 'move', 'ndarray'),
                                         verbose=False)  # print structure below instead

    if existing_dirs is not None and delete_source:
        for i in existing_dirs:
            shutil.rmtree(i)

    if action in ('copy', 'move'):
        to_return = output_dict
    elif action == 'ndarray':
        to_return = _file_paths_dict_to_ndarrays(output_dict, dimensions=2, stack=stack, verbose=verbose)

    if verbose:
        _print_output_dict_structure(output_dict)

    return to_return


def reverse_train_val_test(data, delete_source=True, verbose=True):
    """

    Reverse the action of ``train_val_test`` on a directory.

    :param data: the path to the directory created by ``train_val_test``.
    :type data: ``str``
    :param delete_source: if ``True`` delete the empty train/validation/test directories in ``data`` after the move is
                          complete. Defaults to ``True``.
    :type delete_source: ``bool``
    :param verbose: if ``True``, print the resultant structure. Defaults to ``True``.
    :type verbose: ``bool``

    :Example:

    Initial Directory Structure:

    .. code-block:: bash

        $ tree /path/to/data/images
        ├── ct
        │   ├── ct_1.png
        │   ├── ct_2.png
        │   ├── ct_3.png
        │   ├── ct_4.png
        │   ├── ct_5.png
        │   └── ct_6.png
        └── mri
            ├── mri_1.png
            ├── mri_2.png
            ├── mri_3.png
            ├── mri_4.png
            ├── mri_5.png
            └── mri_6.png


    >>> from biovida.support_tools import train_val_test
    >>> tv = train_val_test(data='/path/to/data/images', train=0.7, validation=0.3, test=None,
    ...                     action='copy', delete_source=True)

    .. code-block:: bash

        $ tree /path/to/data/images
        ├── train
        │   ├── ct
        │   │   ├── ct_4.png
        │   │   ├── ct_5.png
        │   │   ├── ct_3.png
        │   │   └── ct_1.png
        │   └── mri
        │       ├── mri_2.png
        │       ├── mri_1.png
        │       ├── mri_5.png
        │       └── mri_4.png
        └── validation
            ├── ct
            │   ├── ct_2.png
            │   └── ct_6.png
            └── mri
                ├── mri_3.png
                └── mri_6.png

    >>> reverse_train_val_test('/path/to/data/images')

    .. code-block:: bash

        $ tree /path/to/data/images
        ├── ct
        │   ├── ct_1.png
        │   ├── ct_2.png
        │   ├── ct_3.png
        │   ├── ct_4.png
        │   ├── ct_5.png
        │   └── ct_6.png
        └── mri
            ├── mri_1.png
            ├── mri_2.png
            ├── mri_3.png
            ├── mri_4.png
            ├── mri_5.png
            └── mri_6.png

    """
    def subclasses(tvt):
        return [os_join(tvt, c) for c in os.listdir(tvt) if os.path.isdir(os_join(tvt, c))]

    def listdir_full(path):
        return [os_join(path, f) for f in os.listdir(path) if not f.startswith('.')]

    def last_two_path(path):
        return os.sep.join(path.split(os.sep)[-2:])

    tvt = ('train', 'validation', 'test')
    tvt_in_data = [os.path.join(data, d) for d in tvt if os.path.isdir(os.path.join(data, d))]

    if not len(tvt_in_data):
        raise NotADirectoryError("`data` does not contain a 'train', "
                                 "'validation' or 'test' directory.")

    for i in tvt_in_data:
        classes = subclasses(i)
        for c in classes:
            desc = "Processing '{0}'...".format(last_two_path(c))
            for file in tqdm(listdir_full(c), desc=desc, disable=not verbose):
                class_in_data = os_join(data, os.path.basename(c))
                if not os.path.isdir(class_in_data):
                    os.makedirs(class_in_data)
                dst = os_join(class_in_data, os.path.basename(file))
                if not os.path.isfile(dst):
                    shutil.move(src=file, dst=dst)
                else:
                    os.remove(path=file)

    if delete_source:
        for i in tvt_in_data:
            shutil.rmtree(i, ignore_errors=True)

    if verbose:
        print("\nStructure:\n")
        print(list_to_bulletpoints([os.path.basename(i) for i in subclasses(data)]))
